"""
Module providing attribute inference attacks.
"""
from art.attacks.inference.attribute_inference.black_box import AttributeInferenceBlackBox
from art.attacks.inference.attribute_inference.baseline import AttributeInferenceBaseline
from art.attacks.inference.attribute_inference.true_label_baseline import AttributeInferenceBaselineTrueLabel
from art.attacks.inference.attribute_inference.white_box_decision_tree import AttributeInferenceWhiteBoxDecisionTree
from art.attacks.inference.attribute_inference.white_box_lifestyle_decision_tree import (
    AttributeInferenceWhiteBoxLifestyleDecisionTree,
)
from art.attacks.inference.attribute_inference.meminf_based import AttributeInferenceMembership



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements DP-InstaHide training method.

| Paper link: https://arxiv.org/abs/2103.02079

| This training method is dependent to the choice of data augmentation and noise parameters. Consequently, framework
    specific implementations are being provided in ART.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import sys
import time
from typing import List, Optional, Union, Tuple, TYPE_CHECKING

import numpy as np
from tqdm.auto import trange

from art.defences.trainer.trainer import Trainer

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

if TYPE_CHECKING:
    from art.data_generators import DataGenerator
    from art.defences.preprocessor import Preprocessor
    from art.utils import CLASSIFIER_LOSS_GRADIENTS_TYPE, CLIP_VALUES_TYPE

logger = logging.getLogger(__name__)


class DPInstaHideTrainer(Trainer):
    """
    Class performing adversarial training following the DP-InstaHide protocol.

    Uses data augmentation methods in conjunction with some type of additive noise.

    | Paper link: https://arxiv.org/abs/2103.02079
    """

    def __init__(
        self,
        classifier: "CLASSIFIER_LOSS_GRADIENTS_TYPE",
        augmentations: Union["Preprocessor", List["Preprocessor"]],
        noise: Literal["gaussian", "laplacian", "exponential"] = "laplacian",
        loc: Union[int, float] = 0.0,
        scale: Union[int, float] = 0.03,
        clip_values: "CLIP_VALUES_TYPE" = (0.0, 1.0),
    ):
        """
        Create an :class:`.DPInstaHideTrainer` instance.

        :param classifier: The model to train using the protocol.
        :param augmentations: The preprocessing data augmentation defence(s) to be applied.
        :param noise: The type of additive noise to use: 'gaussian' | 'laplacian' | 'exponential'.
        :param loc: The location or mean parameter of the distribution to sample.
        :param scale: The scale or standard deviation parameter of the distribution to sample.
        :param clip_values: Tuple of the form `(min, max)` representing the minimum and maximum values allowed
               for features.
        """
        from art.defences.preprocessor import Preprocessor

        super().__init__(classifier)
        if isinstance(augmentations, Preprocessor):
            self.augmentations = [augmentations]
        else:
            self.augmentations = augmentations
        self.noise = noise
        self.loc = loc
        self.scale = scale
        self.clip_values = clip_values

    def _generate_noise(self, x: np.ndarray) -> np.ndarray:
        if self.noise == "gaussian":
            noise = np.random.normal(loc=self.loc, scale=self.scale, size=x.shape)
        elif self.noise == "laplacian":
            noise = np.random.laplace(loc=self.loc, scale=self.scale, size=x.shape)
        elif self.noise == "exponential":
            noise = np.random.exponential(scale=self.scale, size=x.shape)
        else:
            raise ValueError("The provided noise type is not supported:", self.noise)

        x_noise = x + noise
        x_noise = np.clip(x_noise, self.clip_values[0], self.clip_values[1])

        return x_noise.astype(x.dtype)

    def fit(  # pylint: disable=W0221
        self,
        x: np.ndarray,
        y: np.ndarray,
        validation_data: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        batch_size: int = 128,
        nb_epochs: int = 20,
        **kwargs
    ):
        """
        Train a model adversarially with the DP-InstaHide protocol.
        See class documentation for more information on the exact procedure.

        :param x: Training set.
        :param y: Labels for the training set.
        :param validation_data: Tuple consisting of validation data, (x_val, y_val)
        :param batch_size: Size of batches.
        :param nb_epochs: Number of epochs to use for trainings.
        :param kwargs: Dictionary of framework-specific arguments. These will be passed as such to the `fit` function
               of the target classifier.
        """
        logger.info("Performing adversarial training with DP-InstaHide protocol")

        nb_batches = int(np.ceil(len(x) / batch_size))
        ind = np.arange(len(x))

        logger.info("Adversarial Training DP-InstaHide")

        for i_epoch in trange(nb_epochs, desc="DP-InstaHide training epochs"):
            # shuffle the examples
            np.random.shuffle(ind)
            start_time = time.time()
            train_loss = 0.0
            train_acc = 0.0
            train_n = 0.0

            for batch_id in range(nb_batches):
                # create batch data
                x_batch = x[ind[batch_id * batch_size : min((batch_id + 1) * batch_size, x.shape[0])]]
                y_batch = y[ind[batch_id * batch_size : min((batch_id + 1) * batch_size, x.shape[0])]]

                # augment the data
                x_aug = x_batch.copy()
                y_aug = y_batch.copy()
                for augmentation in self.augmentations:
                    x_aug, y_aug = augmentation(x_aug, y_aug)

                # apply additive noise
                x_aug = self._generate_noise(x_aug)

                # fit batch
                self._classifier.fit(x_aug, y_aug, nb_epochs=1, batch_size=x_aug.shape[0], verbose=0, **kwargs)

                # get metrics
                loss = self._classifier.compute_loss(x_aug, y_aug, reduction="mean")
                output = np.argmax(self.predict(x_batch), axis=1)
                acc = np.sum(output == np.argmax(y_batch, axis=1))
                n = len(x_aug)

                # update metrics
                train_loss += np.sum(loss)
                train_acc += acc
                train_n += n

            train_time = time.time()

            # compute accuracy
            if validation_data is not None:
                (x_test, y_test) = validation_data
                output = np.argmax(self.predict(x_test), axis=1)
                test_loss = self._classifier.compute_loss(x_test, y_test, reduction="mean")
                test_acc = np.mean(output == np.argmax(y_test, axis=1))
                logger.info(
                    "epoch: %s time(s): %.1f, loss(tr): %.4f, acc(tr): %.4f, loss(val): %.4f, acc(val): %.4f",
                    i_epoch,
                    train_time - start_time,
                    train_loss / train_n,
                    train_acc / train_n,
                    test_loss,
                    test_acc,
                )
            else:
                logger.info(
                    "epoch: %s time(s): %.1f, loss: %.4f, acc: %.4f",
                    i_epoch,
                    train_time - start_time,
                    train_loss / train_n,
                    train_acc / train_n,
                )

    def fit_generator(self, generator: "DataGenerator", nb_epochs: int = 20, **kwargs):
        """
        Train a model adversarially with the DP-InstaHide protocol using a data generator.
        See class documentation for more information on the exact procedure.

        :param generator: Data generator.
        :param nb_epochs: Number of epochs to use for trainings.
        :param kwargs: Dictionary of framework-specific arguments. These will be passed as such to the `fit` function
               of the target classifier.
        """
        logger.info("Performing adversarial training with DP-InstaHide protocol")
        size = generator.size
        batch_size = generator.batch_size
        if size is not None:
            nb_batches = int(np.ceil(size / batch_size))
        else:
            raise ValueError("Size is None.")

        logger.info("Adversarial Training DP-InstaHide")

        for i_epoch in trange(nb_epochs, desc="DP-InstaHide training epochs"):
            start_time = time.time()
            train_loss = 0.0
            train_acc = 0.0
            train_n = 0.0

            for _ in range(nb_batches):
                # create batch data
                x_batch, y_batch = generator.get_batch()

                # augment the data
                x_aug = x_batch.copy()
                y_aug = y_batch.copy()
                for augmentation in self.augmentations:
                    x_aug, y_aug = augmentation(x_aug, y_aug)

                # apply additive noise
                x_aug = self._generate_noise(x_aug)

                # fit batch
                self._classifier.fit(x_aug, y_aug, nb_epochs=1, batch_size=x_aug.shape[0], verbose=0, **kwargs)

                # get metrics
                loss = self._classifier.compute_loss(x_aug, y_aug, reduction="mean")
                output = np.argmax(self.predict(x_batch), axis=1)
                acc = np.sum(output == np.argmax(y_batch, axis=1))
                n = len(x_aug)

                # update metrics
                train_loss += np.sum(loss)
                train_acc += acc
                train_n += n

            train_time = time.time()

            # compute accuracy
            logger.info(
                "epoch: %s time(s): %.1f, loss: %.4f, acc: %.4f",
                i_epoch,
                train_time - start_time,
                train_loss / train_n,
                train_acc / train_n,
            )

    def predict(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """
        Perform prediction using the adversarially trained classifier.

        :param x: Input samples.
        :param kwargs: Other parameters to be passed on to the `predict` function of the classifier.
        :return: Predictions for test set.
        """
        return self._classifier.predict(x, **kwargs)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a wrapper for video compression defence with FFmpeg.

| Please keep in mind the limitations of defences. For details on how to evaluate classifier security in general,
    see https://arxiv.org/abs/1902.06705.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
from tempfile import TemporaryDirectory
from typing import Optional, Tuple
import warnings

import numpy as np
from tqdm.auto import tqdm

from art import config
from art.defences.preprocessor.preprocessor import Preprocessor

logger = logging.getLogger(__name__)


class VideoCompression(Preprocessor):
    """
    Implement FFmpeg wrapper for video compression defence based on H.264/MPEG-4 AVC.

    Video compression uses H.264 video encoding. The video quality is controlled with the constant rate factor
    parameter. More information on the constant rate factor: https://trac.ffmpeg.org/wiki/Encode/H.264.
    """

    params = ["video_format", "constant_rate_factor", "channels_first", "verbose"]

    def __init__(
        self,
        *,
        video_format: str,
        constant_rate_factor: int = 28,
        channels_first: bool = False,
        apply_fit: bool = False,
        apply_predict: bool = True,
        verbose: bool = False,
    ):
        """
        Create an instance of VideoCompression.

        :param video_format: Specify one of supported video file extensions, e.g. `avi`, `mp4` or `mkv`.
        :param constant_rate_factor: Specify constant rate factor (range 0 to 51, where 0 is lossless).
        :param channels_first: Set channels first or last.
        :param apply_fit: True if applied during fitting/training.
        :param apply_predict: True if applied during predicting.
        :param verbose: Show progress bars.
        """
        super().__init__(is_fitted=True, apply_fit=apply_fit, apply_predict=apply_predict)
        self.video_format = video_format
        self.constant_rate_factor = constant_rate_factor
        self.channels_first = channels_first
        self.verbose = verbose
        self._check_params()

    def __call__(self, x: np.ndarray, y: Optional[np.ndarray] = None) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """
        Apply video compression to sample `x`.

        :param x: Sample to compress of shape NCFHW or NFHWC. `x` values are expected to be either in range [0, 1] or
                  [0, 255].
        :param y: Labels of the sample `x`. This function does not affect them in any way.
        :return: Compressed sample.
        """

        def compress_video(x: np.ndarray, video_format: str, constant_rate_factor: int, dir_: str = ""):
            """
            Apply video compression to video input of shape (frames, height, width, channel).
            """
            import ffmpeg

            video_path = os.path.join(dir_, f"tmp_video.{video_format}")
            _, height, width, _ = x.shape

            if (height % 2) != 0 or (width % 2) != 0:
                warnings.warn("Codec might require even number of pixels in height and width.")

            # numpy to local video file
            process = (
                ffmpeg.input("pipe:", format="rawvideo", pix_fmt="rgb24", s=f"{width}x{height}")
                .output(video_path, pix_fmt="yuv420p", vcodec="libx264", crf=constant_rate_factor)
                .overwrite_output()
                .run_async(pipe_stdin=True, quiet=True)
            )
            process.stdin.write(x.flatten().astype(np.uint8).tobytes())
            process.stdin.close()
            process.wait()

            # local video file to numpy
            stdout, _ = (
                ffmpeg.input(video_path)
                .output("pipe:", format="rawvideo", pix_fmt="rgb24")
                .run(capture_stdout=True, quiet=True)
            )
            return np.frombuffer(stdout, np.uint8).reshape(x.shape)

        if x.ndim != 5:
            raise ValueError("Video compression can only be applied to spatio-temporal data.")

        if self.channels_first:
            x = np.transpose(x, (0, 2, 3, 4, 1))

        # apply video compression per video item
        scale = 1
        if x.min() >= 0 and x.max() <= 1.0:
            scale = 255

        x_compressed = x.copy()
        with TemporaryDirectory(dir=config.ART_DATA_PATH) as tmp_dir:
            for i, x_i in enumerate(tqdm(x, desc="Video compression", disable=not self.verbose)):
                x_i *= scale
                x_compressed[i] = compress_video(x_i, self.video_format, self.constant_rate_factor, dir_=tmp_dir)

        x_compressed = x_compressed / scale
        x_compressed = x_compressed.astype(x.dtype)

        if self.channels_first:
            x_compressed = np.transpose(x_compressed, (0, 4, 1, 2, 3))

        return x_compressed, y

    def _check_params(self) -> None:
        if not (isinstance(self.constant_rate_factor, int) and 0 <= self.constant_rate_factor < 52):
            raise ValueError("Constant rate factor must be an integer in the range [0, 51].")

        if not isinstance(self.verbose, bool):
            raise ValueError("The argument `verbose` has to be of type bool.")



import logging
import os
from os import listdir, path
import pickle
import tempfile

import keras
import numpy as np
import pytest
from tensorflow.keras.callbacks import LearningRateScheduler

from art.defences.preprocessor import FeatureSqueezing, JpegCompression, SpatialSmoothing
from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.skip_framework("non_dl_frameworks")
def test_get_activations(art_warning, get_default_mnist_subset, framework, image_dl_estimator):
    try:
        classifier, _ = image_dl_estimator(from_logits=True)

        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        batch_size = 128
        for i, name in enumerate(classifier.layer_names):
            activation_i = classifier.get_activations(x_test_mnist, i, batch_size=batch_size)
            activation_name = classifier.get_activations(x_test_mnist, name, batch_size=batch_size)
            np.testing.assert_array_equal(activation_name, activation_i)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
def test_loss_gradient_with_wildcard(art_warning, image_dl_estimator):
    try:
        classifier, _ = image_dl_estimator(wildcard=True)
        shapes = [(1, 10, 1), (1, 20, 1)]
        for shape in shapes:
            x = np.random.normal(size=shape)
            loss_gradient = classifier.loss_gradient(x, y=[1])
            assert loss_gradient.shape == shape

            class_gradient = classifier.class_gradient(x, 0)
            assert class_gradient[0].shape == shape

    except ARTTestException as e:
        art_warning(e)


# Note: because mxnet only supports 1 concurrent version of a model if we fit that model, all expected values will
# change for all other tests using that fitted model
@pytest.mark.skip_framework("mxnet", "non_dl_frameworks")
def test_fit(art_warning, get_default_mnist_subset, default_batch_size, image_dl_estimator):
    try:
        (x_train_mnist, y_train_mnist), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        labels = np.argmax(y_test_mnist, axis=1)
        classifier, sess = image_dl_estimator(from_logits=True)

        accuracy = np.sum(np.argmax(classifier.predict(x_test_mnist), axis=1) == labels) / x_test_mnist.shape[0]
        np.testing.assert_array_almost_equal(accuracy, 0.32, decimal=0.06)

        classifier.fit(x_train_mnist, y_train_mnist, batch_size=default_batch_size, nb_epochs=2)
        accuracy_2 = np.sum(np.argmax(classifier.predict(x_test_mnist), axis=1) == labels) / x_test_mnist.shape[0]
        np.testing.assert_array_almost_equal(accuracy_2, 0.73, decimal=0.06)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
@pytest.mark.skipif(keras.__version__.startswith("2.2"), reason="requires Keras 2.3.0 or higher")
def test_predict(
    art_warning, framework, get_default_mnist_subset, image_dl_estimator, expected_values, store_expected_values
):
    try:
        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        classifier, _ = image_dl_estimator(from_logits=True)

        y_predicted = classifier.predict(x_test_mnist[0:1])
        np.testing.assert_array_almost_equal(y_predicted, expected_values(), decimal=4)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
def test_shapes(art_warning, get_default_mnist_subset, image_dl_estimator):
    try:
        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset
        classifier, sess = image_dl_estimator(from_logits=True)

        predictions = classifier.predict(x_test_mnist)
        assert predictions.shape == y_test_mnist.shape

        assert classifier.nb_classes == 10

        class_gradients = classifier.class_gradient(x_test_mnist[:11])
        assert class_gradients.shape == tuple([11, 10] + list(x_test_mnist[1].shape))

        loss_gradients = classifier.loss_gradient(x_test_mnist[:11], y_test_mnist[:11])
        assert loss_gradients.shape == x_test_mnist[:11].shape

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
@pytest.mark.parametrize("from_logits", [True, False])
@pytest.mark.parametrize(
    "loss_name",
    ["categorical_crossentropy", "categorical_hinge", "sparse_categorical_crossentropy", "kullback_leibler_divergence"],
)
def test_loss_functions(
    art_warning,
    image_dl_estimator,
    get_default_mnist_subset,
    loss_name,
    supported_losses_proba,
    supported_losses_logit,
    store_expected_values,
    supported_losses_types,
    from_logits,
    expected_values,
):
    # prediction and class_gradient should be independent of logits/probabilities and of loss function

    try:
        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        if from_logits:
            supported_losses = supported_losses_logit()
        else:
            supported_losses = supported_losses_proba()

        for loss_type in supported_losses_types():
            (y_test_pred_exp, class_gradient_exp, loss_grad_exp) = expected_values()
            # store_expected_values(expected_values)

            if loss_name + "_" + loss_type in supported_losses:
                classifier, _ = image_dl_estimator(loss_name=loss_name, loss_type=loss_type, from_logits=from_logits)

                y_test_pred_exp = np.argmax(classifier.predict(x=x_test_mnist), axis=1)
                np.testing.assert_array_equal(y_test_pred_exp, y_test_pred_exp)

                class_gradient = classifier.class_gradient(x_test_mnist, label=5)
                # np.testing.assert_array_almost_equal(class_gradient[99, 0, 14, :, 0], class_gradient_prob_exp)
                np.testing.assert_array_almost_equal(class_gradient[99, 0, 14, :, 0], class_gradient_exp)

                loss_gradient_value = classifier.loss_gradient(x=x_test_mnist, y=y_test_mnist)
                np.testing.assert_array_almost_equal(loss_gradient_value[99, 14, :, 0], loss_grad_exp[loss_name])

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
def test_pickle(art_warning, image_dl_estimator, image_dl_estimator_defended, tmp_path):
    try:
        full_path = os.path.join(tmp_path, "my_classifier.p")

        classifier, _ = image_dl_estimator(functional=True)
        with open(full_path, "wb") as save_file:
            pickle.dump(classifier, save_file)

        with open(full_path, "rb") as load_file:
            loaded = pickle.load(load_file)

        assert (classifier._clip_values == loaded._clip_values).all()
        assert classifier._channels_first == loaded._channels_first
        if hasattr(classifier, "_use_logits"):
            assert classifier._use_logits == loaded._use_logits
        if hasattr(classifier, "_input_layer"):
            assert classifier._input_layer == loaded._input_layer
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks", "pytorch")
def test_functional_model(art_warning, image_dl_estimator):
    try:
        # Need to update the functional_model code to produce a model with more than one input and output layers...
        classifier, _ = image_dl_estimator(functional=True, input_layer=1, output_layer=1)
        assert "input1" in classifier._input.name
        assert "output1" in classifier._output.name

        classifier, _ = image_dl_estimator(functional=True, input_layer=0, output_layer=0)
        assert "input0" in classifier._input.name
        assert "output0" in classifier._output.name
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("mxnet", "tensorflow", "pytorch", "non_dl_frameworks")
def test_fit_kwargs(art_warning, image_dl_estimator, get_default_mnist_subset, default_batch_size):
    try:
        (x_train_mnist, y_train_mnist), (_, _) = get_default_mnist_subset

        def get_lr(_):
            return 0.01

        # Test a valid callback
        classifier, _ = image_dl_estimator(from_logits=True)
        kwargs = {"callbacks": [LearningRateScheduler(get_lr)]}
        classifier.fit(x_train_mnist, y_train_mnist, batch_size=default_batch_size, nb_epochs=1, **kwargs)

        # Test failure for invalid parameters
        kwargs = {"epochs": 1}
        with pytest.raises(TypeError) as exception:
            classifier.fit(x_train_mnist, y_train_mnist, batch_size=default_batch_size, nb_epochs=1, **kwargs)

        assert "multiple values for keyword argument" in str(exception)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
def test_defences_predict(art_warning, get_default_mnist_subset, image_dl_estimator_defended, image_dl_estimator):
    try:
        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        classifier, _ = image_dl_estimator()
        y_check_clean = classifier.predict(x_test_mnist)
        clip_values = (0, 1)

        classifier_defended, _ = image_dl_estimator_defended(defenses=["FeatureSqueezing"])
        assert len(classifier_defended.preprocessing_defences) == 1
        y_defended = classifier_defended.predict(x_test_mnist)
        fs = FeatureSqueezing(clip_values=clip_values, bit_depth=2)
        x_test_defense, _ = fs(x_test_mnist, y_test_mnist)
        y_check = classifier.predict(x_test_defense)
        np.testing.assert_array_almost_equal(y_defended, y_check, decimal=4)
        np.testing.assert_raises(AssertionError, np.testing.assert_array_equal, y_check, y_check_clean)

        classifier_defended, _ = image_dl_estimator_defended(defenses=["JpegCompression"])
        assert len(classifier_defended.preprocessing_defences) == 1
        y_defended = classifier_defended.predict(x_test_mnist)
        jpeg = JpegCompression(
            clip_values=clip_values, apply_predict=True, channels_first=classifier_defended.channels_first
        )
        x_test_defense, _ = jpeg(x_test_mnist, y_test_mnist)
        y_check = classifier.predict(x_test_defense)
        np.testing.assert_array_almost_equal(y_defended, y_check, decimal=4)
        np.testing.assert_raises(AssertionError, np.testing.assert_array_equal, y_check, y_check_clean)

        classifier_defended, _ = image_dl_estimator_defended(defenses=["SpatialSmoothing"])
        assert len(classifier_defended.preprocessing_defences) == 1
        y_defended = classifier_defended.predict(x_test_mnist)
        smooth = SpatialSmoothing(channels_first=classifier_defended.channels_first)
        x_test_defense, _ = smooth(x_test_mnist, y_test_mnist)
        y_check = classifier.predict(x_test_defense)
        np.testing.assert_array_almost_equal(y_defended, y_check, decimal=4)
        np.testing.assert_raises(AssertionError, np.testing.assert_array_equal, y_check, y_check_clean)

    except ARTTestException as e:
        art_warning(e)


# Note: because mxnet only supports 1 concurrent version of a model if we fit that model, all expected values will
# change for all other tests using that fitted model
@pytest.mark.skip_framework("mxnet", "non_dl_frameworks")
def test_fit_image_generator(
    art_warning, framework, image_dl_estimator, image_data_generator, get_default_mnist_subset
):
    try:
        classifier, sess = image_dl_estimator(from_logits=True)
        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        true_class = np.argmax(y_test_mnist, axis=1)

        predictions = classifier.predict(x_test_mnist)
        prediction_class = np.argmax(predictions, axis=1)
        pre_fit_accuracy = np.sum(prediction_class == true_class) / x_test_mnist.shape[0]

        assert pre_fit_accuracy == pytest.approx(0.32, abs=0.01)

        data_gen = image_data_generator(sess=sess)
        classifier.fit_generator(generator=data_gen, nb_epochs=2)
        predictions = classifier.predict(x_test_mnist)
        prediction_class = np.argmax(predictions, axis=1)
        post_fit_accuracy = np.sum(prediction_class == true_class) / x_test_mnist.shape[0]

        np.testing.assert_array_almost_equal(post_fit_accuracy, 0.68, decimal=0.06)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
@pytest.mark.skipif(keras.__version__.startswith("2.2"), reason="requires Keras 2.3.0 or higher")
def test_loss_gradient(
    art_warning,
    framework,
    get_default_mnist_subset,
    image_dl_estimator,
    expected_values,
    mnist_shape,
    store_expected_values,
):
    try:
        (expected_gradients_1, expected_gradients_2) = expected_values()

        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset
        classifier, _ = image_dl_estimator(from_logits=True)

        gradients = classifier.loss_gradient(x_test_mnist, y_test_mnist)

        assert gradients.shape == (x_test_mnist.shape[0],) + mnist_shape

        if mnist_shape[0] == 1:
            sub_gradients = gradients[0, 0, :, 14]
        else:
            sub_gradients = gradients[0, :, 14, 0]

        np.testing.assert_array_almost_equal(
            sub_gradients,
            expected_gradients_1[0],
            decimal=expected_gradients_1[1],
        )

        if mnist_shape[0] == 1:
            sub_gradients = gradients[0, 0, 14, :]
        else:
            sub_gradients = gradients[0, 14, :, 0]

        np.testing.assert_array_almost_equal(
            sub_gradients,
            expected_gradients_2[0],
            decimal=expected_gradients_2[1],
        )
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
def test_nb_classes(art_warning, image_dl_estimator):
    try:
        classifier, _ = image_dl_estimator(from_logits=True)
        assert classifier.nb_classes == 10
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
def test_input_shape(art_warning, image_dl_estimator, mnist_shape):
    try:
        classifier, _ = image_dl_estimator(from_logits=True)
        assert classifier.input_shape == mnist_shape
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("tensorflow2", "non_dl_frameworks")
def test_save_1(art_warning, image_dl_estimator):
    try:
        classifier, _ = image_dl_estimator(from_logits=True)
        t_file = tempfile.NamedTemporaryFile()
        model_path = t_file.name
        t_file.close()
        filename = "model_to_save"
        classifier.save(filename, path=model_path)

        assert path.exists(model_path)

        created_model = False

        for file in listdir(model_path):
            if filename in file:
                created_model = True
        assert created_model

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("tensorflow", "non_dl_frameworks")
def test_save_2(art_warning, image_dl_estimator, get_default_mnist_subset, tmp_path):
    try:
        classifier, _ = image_dl_estimator()
        (x_train_mnist, y_train_mnist), (_, _) = get_default_mnist_subset
        classifier.fit(x_train_mnist, y_train_mnist, batch_size=128, nb_epochs=2)
        full_path = tmp_path / "sub"
        full_path.mkdir()

        assert not os.listdir(full_path._str)
        classifier.save("modelFile", path=full_path._str)
        assert os.listdir(full_path._str)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
def test_repr(art_warning, image_dl_estimator, framework, expected_values, store_expected_values):
    try:
        classifier, _ = image_dl_estimator(from_logits=True)
        repr_ = repr(classifier)
        for message in expected_values():
            assert message in repr_, "{0}: was not contained within repr".format(message)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks")
@pytest.mark.skipif(keras.__version__.startswith("2.2"), reason="requires Keras 2.3.0 or higher")
def test_class_gradient(
    art_warning,
    framework,
    image_dl_estimator,
    get_default_mnist_subset,
    mnist_shape,
    store_expected_values,
    expected_values,
):
    try:
        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        classifier, _ = image_dl_estimator(from_logits=True)

        (
            grad_1_all_labels,
            grad_2_all_labels,
            grad_1_label5,
            grad_2_label5,
            grad_1_labelArray,
            grad_2_labelArray,
            labels_list,
        ) = expected_values()

        labels = np.array(labels_list, dtype=object)

        # TODO we should consider checking channel independent columns to make this test truly framework independent
        def get_gradient1_column(gradients):
            if mnist_shape[0] == 1:
                return gradients[0, 5, 0, 14, :]  # expected_gradients_1_all_labels
            else:
                return gradients[0, 5, 14, :, 0]

        def get_gradient2_column(gradients):
            if mnist_shape[0] == 1:
                return gradients[0, 5, 0, :, 14]  # expected_gradients_2_all_labels
            else:
                return gradients[0, 5, :, 14, 0]

        def get_gradient3_column(gradients):
            if mnist_shape[0] == 1:
                return gradients[0, 0, 0, 14, :]  # expected_gradients_1_label5
            else:
                return gradients[0, 0, 14, :, 0]

        def get_gradient4_column(gradients):
            if mnist_shape[0] == 1:
                return gradients[0, 0, 0, :, 14]  # expected_gradients_2_all_labels
            else:
                return gradients[0, 0, :, 14, 0]

        # Test all gradients label
        gradients = classifier.class_gradient(x_test_mnist)

        new_shape = (
            x_test_mnist.shape[0],
            10,
        ) + mnist_shape
        assert gradients.shape == new_shape

        sub_gradients2 = get_gradient2_column(gradients)
        if framework != "mxnet":
            np.testing.assert_array_almost_equal(
                sub_gradients2,
                grad_2_all_labels[0],
                decimal=4,
            )

        # Test 1 gradient label = 5
        gradients = classifier.class_gradient(x_test_mnist, label=5)

        assert (
            gradients.shape
            == (
                x_test_mnist.shape[0],
                1,
            )
            + mnist_shape
        )

        sub_gradients2 = get_gradient3_column(gradients)
        if framework != "mxnet":
            np.testing.assert_array_almost_equal(
                sub_gradients2,
                grad_1_label5[0],
                decimal=4,
            )

        sub_gradients4 = get_gradient4_column(gradients)
        if framework != "mxnet":
            np.testing.assert_array_almost_equal(
                sub_gradients4,
                grad_2_label5[0],
                decimal=4,
            )

        # # Test a set of gradients label = array
        gradients = classifier.class_gradient(x_test_mnist, label=labels)

        new_shape = (
            x_test_mnist.shape[0],
            1,
        ) + mnist_shape
        assert gradients.shape == new_shape

        sub_gradients5 = get_gradient3_column(gradients)
        if framework != "mxnet":
            np.testing.assert_array_almost_equal(
                sub_gradients5,
                grad_1_labelArray[0],
                decimal=4,
            )

        sub_gradients6 = get_gradient4_column(gradients)
        if framework != "mxnet":
            np.testing.assert_array_almost_equal(
                sub_gradients6,
                grad_2_labelArray[0],
                decimal=4,
            )

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("mxnet", "non_dl_frameworks")
def test_compute_loss(
    art_warning,
    framework,
    image_dl_estimator,
    get_default_mnist_subset,
    mnist_shape,
    store_expected_values,
    expected_values,
):
    try:
        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset
        classifier, _ = image_dl_estimator(from_logits=True)
        expected_loss = expected_values()
        computed_loss = float(classifier.compute_loss(x=x_test_mnist, y=y_test_mnist, reduction="sum"))

        assert pytest.approx(computed_loss, rel=0.01) in expected_loss

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("keras", "kerastf", "mxnet", "non_dl_frameworks")
def test_clone_for_refitting(
    art_warning,
    image_dl_estimator,
):
    try:
        classifier, _ = image_dl_estimator(functional=True)
        _ = classifier.clone_for_refitting()

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
A unittest class for testing the subset scanning detector.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pytest

import numpy as np

from art.attacks.evasion.fast_gradient import FastGradientMethod
from art.defences.detector.evasion import SubsetScanningDetector

from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.only_with_platform("keras", "kerastf", "tensorflow2", "pytorch")
def test_subsetscannning_detector_scan_clean(art_warning, get_default_mnist_subset, image_dl_estimator):
    (x_train, _), (x_test, _) = get_default_mnist_subset
    classifier, _ = image_dl_estimator()

    # Data for detector
    bgd_data = x_train
    clean_data = x_test

    try:
        detector = SubsetScanningDetector(classifier, bgd_data=bgd_data, layer=1)
        _, _, dpwr = detector.scan(clean_x=clean_data, adv_x=clean_data)
        assert dpwr > 0
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("keras", "kerastf", "tensorflow2", "pytorch")
def test_subsetscannning_detector_scan_adv(art_warning, get_default_mnist_subset, image_dl_estimator):
    (x_train, _), (x_test, _) = get_default_mnist_subset
    classifier, _ = image_dl_estimator()

    # Generate adversarial samples
    attacker = FastGradientMethod(classifier, eps=0.5)
    x_test_adv = attacker.generate(x_test)

    # Data for detector
    bgd_data = x_train
    clean_data = x_test
    adv_data = x_test_adv

    try:
        detector = SubsetScanningDetector(classifier, bgd_data=bgd_data, layer=1)
        _, _, dpwr = detector.scan(clean_x=clean_data, adv_x=adv_data)
        assert dpwr > 0
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("keras", "kerastf", "tensorflow2", "pytorch")
def test_subsetscannning_detector_scan_size(art_warning, get_default_mnist_subset, image_dl_estimator):
    (x_train, _), (x_test, _) = get_default_mnist_subset
    classifier, _ = image_dl_estimator()

    # Generate adversarial samples
    attacker = FastGradientMethod(classifier, eps=0.5)
    x_test_adv = attacker.generate(x_test)

    # Compile training data for detector:
    bgd_data = x_train
    clean_data = x_test
    adv_data = np.concatenate((x_test, x_test_adv), axis=0)

    try:
        detector = SubsetScanningDetector(classifier, bgd_data=bgd_data, layer=1)
        _, _, dpwr = detector.scan(clean_x=clean_data, adv_x=adv_data, clean_size=85, adv_size=15)
        assert dpwr > 0
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("keras", "kerastf", "tensorflow2", "pytorch")
def test_subsetscannning_detector_detect(art_warning, get_default_mnist_subset, image_dl_estimator):
    (x_train, _), (x_test, _) = get_default_mnist_subset
    classifier, _ = image_dl_estimator()

    # Generate adversarial samples
    attacker = FastGradientMethod(classifier, eps=0.5)
    x_test_adv = attacker.generate(x_test)

    # Data for detector
    bgd_data = x_train
    adv_data = x_test_adv

    try:
        detector = SubsetScanningDetector(classifier, bgd_data=bgd_data, layer=1)
        _, is_adversarial = detector.detect(adv_data)
        assert len(is_adversarial) == len(adv_data)
    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements (De)Randomized Smoothing for Certifiable Defense against Patch Attacks

| Paper link: https://arxiv.org/abs/2002.10733
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from abc import ABC, abstractmethod
from typing import Optional, Union, TYPE_CHECKING
import random

import numpy as np

if TYPE_CHECKING:
    from art.utils import ABLATOR_TYPE


class DeRandomizedSmoothingMixin(ABC):
    """
    Implementation of (De)Randomized Smoothing applied to classifier predictions as introduced
    in Levine et al. (2020).

    | Paper link: https://arxiv.org/abs/2002.10733
    """

    def __init__(
        self,
        ablation_type: str,
        ablation_size: int,
        threshold: float,
        logits: bool,
        channels_first: bool,
        *args,
        **kwargs,
    ) -> None:
        """
        Create a derandomized smoothing wrapper.

        :param ablation_type: The type of ablations to perform. Currently must be either "column", "row", or "block"
        :param ablation_size: Size of the retained image patch.
                              An int specifying the width of the column for column ablation
                              Or an int specifying the height/width of a square for block ablation
        :param threshold: The minimum threshold to count a prediction.
        :param logits: if the model returns logits or normalized probabilities
        :param channels_first: If the channels are first or last.
        """
        super().__init__(*args, **kwargs)  # type: ignore
        self.ablation_type = ablation_type
        self.logits = logits
        self.threshold = threshold
        self._channels_first = channels_first
        if TYPE_CHECKING:
            self.ablator: ABLATOR_TYPE  # pylint: disable=used-before-assignment

        if self.ablation_type in {"column", "row"}:
            row_ablation_mode = self.ablation_type == "row"
            self.ablator = ColumnAblator(
                ablation_size=ablation_size, channels_first=self._channels_first, row_ablation_mode=row_ablation_mode
            )
        elif self.ablation_type == "block":
            self.ablator = BlockAblator(ablation_size=ablation_size, channels_first=self._channels_first)
        else:
            raise ValueError("Ablation type not supported. Must be either column or block")

    def _predict_classifier(self, x: np.ndarray, batch_size: int, training_mode: bool, **kwargs) -> np.ndarray:
        """
        Perform prediction for a batch of inputs.

        :param x: Input samples.
        :param batch_size: Size of batches.
        :param training_mode: `True` for model set to training mode and `'False` for model set to evaluation mode.
        :return: Array of predictions of shape `(nb_inputs, nb_classes)`.
        """
        raise NotImplementedError

    def predict(self, x: np.ndarray, batch_size: int = 128, training_mode: bool = False, **kwargs) -> np.ndarray:
        """
        Performs cumulative predictions over every ablation location

        :param x: Unablated image
        :param batch_size: the batch size for the prediction
        :param training_mode: if to run the classifier in training mode
        :return: cumulative predictions after sweeping over all the ablation configurations.
        """
        if self._channels_first:
            columns_in_data = x.shape[-1]
            rows_in_data = x.shape[-2]
        else:
            columns_in_data = x.shape[-2]
            rows_in_data = x.shape[-3]

        if self.ablation_type in {"column", "row"}:
            if self.ablation_type == "column":
                ablate_over_range = columns_in_data
            else:
                # image will be transposed, so loop over the number of rows
                ablate_over_range = rows_in_data

            for ablation_start in range(ablate_over_range):
                ablated_x = self.ablator.forward(np.copy(x), column_pos=ablation_start)
                if ablation_start == 0:
                    preds = self._predict_classifier(
                        ablated_x, batch_size=batch_size, training_mode=training_mode, **kwargs
                    )
                else:
                    preds += self._predict_classifier(
                        ablated_x, batch_size=batch_size, training_mode=training_mode, **kwargs
                    )
        elif self.ablation_type == "block":
            for xcorner in range(rows_in_data):
                for ycorner in range(columns_in_data):
                    ablated_x = self.ablator.forward(np.copy(x), row_pos=xcorner, column_pos=ycorner)
                    if ycorner == 0 and xcorner == 0:
                        preds = self._predict_classifier(
                            ablated_x, batch_size=batch_size, training_mode=training_mode, **kwargs
                        )
                    else:
                        preds += self._predict_classifier(
                            ablated_x, batch_size=batch_size, training_mode=training_mode, **kwargs
                        )
        return preds


class BaseAblator(ABC):
    """
    Base class defining the methods used for the ablators.
    """

    @abstractmethod
    def __call__(
        self, x: np.ndarray, column_pos: Optional[Union[int, list]] = None, row_pos: Optional[Union[int, list]] = None
    ) -> np.ndarray:
        """
        Ablate the image x at location specified by "column_pos" for the case of column ablation or at the location
        specified by "column_pos" and "row_pos" in the case of block ablation.

        :param x: input image.
        :param column_pos: column position to specify where to retain the image
        :param row_pos: row position to specify where to retain the image. Not used for ablation type "column".
        """
        raise NotImplementedError

    @abstractmethod
    def certify(self, preds: np.ndarray, size_to_certify: int):
        """
        Checks if based on the predictions supplied the classifications over the ablated datapoints result in a
        certified prediction against a patch attack of size size_to_certify.

        :param preds: The cumulative predictions of the classifier over the ablation locations.
        :param size_to_certify: The size of the patch to check against.
        """
        raise NotImplementedError

    @abstractmethod
    def ablate(self, x: np.ndarray, column_pos: int, row_pos: int) -> np.ndarray:
        """
        Ablate the image x at location specified by "column_pos" for the case of column ablation or at the location
        specified by "column_pos" and "row_pos" in the case of block ablation.

        :param x: input image.
        :param column_pos: column position to specify where to retain the image
        :param row_pos: row position to specify where to retain the image. Not used for ablation type "column".
        """
        raise NotImplementedError

    @abstractmethod
    def forward(
        self, x: np.ndarray, column_pos: Optional[Union[int, list]] = None, row_pos: Optional[Union[int, list]] = None
    ) -> np.ndarray:
        """
        Ablate batch of data at locations specified by column_pos and row_pos

        :param x: input image.
        :param column_pos: column position to specify where to retain the image
        :param row_pos: row position to specify where to retain the image. Not used for ablation type "column".
        """
        raise NotImplementedError


class ColumnAblator(BaseAblator):
    """
    Implements the functionality for albating the image, and retaining only a column
    """

    def __init__(self, ablation_size: int, channels_first: bool, row_ablation_mode: bool = False):
        """
        Creates an ablator which will retain columns by default, or rows if operating in row_ablation_mode

        :param ablation_size: Size of the column (or row if running in row_ablation_mode) to retain.
        :param channels_first: If the input data will be in channels first or channels last format.
        :param row_ablation_mode: if True then the ablator will function by retaining rows rather than columns.
        """
        super().__init__()
        self.ablation_size = ablation_size
        self.channels_first = channels_first
        self.row_ablation_mode = row_ablation_mode

    def __call__(
        self, x: np.ndarray, column_pos: Optional[Union[int, list]] = None, row_pos: Optional[Union[int, list]] = None
    ) -> np.ndarray:
        """
        Performs ablation on the input x. If no column_pos is specified a random location will be selected.

        :param x: input image.
        :param column_pos: int indicating the start column to retain across all samples in the batch
                           or list of ints of length equal to the number of samples to have a different
                           column retained per sample. If not supplied a random location will be selected.
                           NB, if row_ablation_mode is true then this will be used to act on the rows through
                           transposing the image.
        :param row_pos: Unused
        :return: ablated image keeping only a column.
        """
        return self.forward(x=x, column_pos=column_pos)

    def certify(self, preds: np.ndarray, size_to_certify: int) -> np.ndarray:
        """
        Checks if based on the predictions supplied the classifications over the ablated datapoints result in a
        certified prediction against a patch attack of size size_to_certify.

        :param preds: The cumulative predictions of the classifier over the ablation locations.
        :param size_to_certify: The size of the patch to check against.
        :return: Array of bools indicating if a point is certified against the given patch dimensions.
        """
        indices = np.argsort(-preds, axis=1, kind="stable")
        values = np.take_along_axis(np.copy(preds), indices, axis=1)

        num_affected_classifications = size_to_certify + self.ablation_size - 1

        margin = values[:, 0] - values[:, 1]

        certs = margin > 2 * num_affected_classifications
        tie_break_certs = (margin == 2 * num_affected_classifications) & (indices[:, 0] < indices[:, 1])
        return np.logical_or(certs, tie_break_certs)

    def ablate(self, x: np.ndarray, column_pos: int, row_pos=None) -> np.ndarray:
        """
        Ablates the image only retaining a column starting at "pos" of width "self.ablation_size"

        :param x: input image.
        :param column_pos: location to start the retained column. NB, if row_ablation_mode is true then this will
                           be used to act on the rows through transposing the image.
        :param row_pos: Unused.
        :return: ablated image keeping only a column.
        """
        if self.row_ablation_mode:
            x = np.transpose(x, (0, 1, 3, 2))

        k = self.ablation_size
        num_of_image_columns = x.shape[-1]

        if column_pos + k > num_of_image_columns:
            start_of_ablation = column_pos + k - num_of_image_columns
            x[:, :, :, start_of_ablation:column_pos] = 0.0
        else:
            x[:, :, :, :column_pos] = 0.0
            x[:, :, :, column_pos + k :] = 0.0

        if self.row_ablation_mode:
            x = np.transpose(x, (0, 1, 3, 2))

        return x

    def forward(
        self, x: np.ndarray, column_pos: Optional[Union[int, list]] = None, row_pos: Optional[Union[int, list]] = None
    ) -> np.ndarray:
        """
        Performs ablation on the input x. If no column_pos is specified a random location will be selected.

        :param x: input batch.
        :param column_pos: int indicating the start column to retain across all samples in the batch
                           or list of ints of length equal to the number of samples to have a different
                           column retained per sample. If not supplied a random location will be selected.
                           NB, if row_ablation_mode is true then this will be used to act on the rows through
                           transposing the image.
        :param row_pos: Unused.
        :return: Batch ablated according to the locations in column_pos. Data is channel extended to indicate to a
                 model if a position is ablated.
        """
        if not self.channels_first:
            x = np.transpose(x, (0, 3, 1, 2))

        x = np.concatenate([x, 1.0 - x], axis=1)

        if column_pos is None:
            column_pos = random.randint(0, x.shape[3])

        if isinstance(column_pos, list):
            assert len(column_pos) == len(x)
            for i, pos in enumerate(column_pos):
                x[i : i + 1] = self.ablate(x[i : i + 1], pos)
        else:
            x = self.ablate(x, column_pos)

        if not self.channels_first:
            x = np.transpose(x, (0, 2, 3, 1))

        return x


class BlockAblator(BaseAblator):
    """
    Implements the functionality for albating the image, and retaining only a block
    """

    def __init__(self, ablation_size: int, channels_first: bool):
        """
        Creates an ablator which will retain blocks of the input data.

        :param ablation_size: Size of the column (or row if running in row_ablation_mode) to retain.
        :param channels_first: If the input data will be in channels first or channels last format.
        """
        super().__init__()
        self.ablation_size = ablation_size
        self.channels_first = channels_first

    def __call__(
        self, x: np.ndarray, column_pos: Optional[Union[int, list]] = None, row_pos: Optional[Union[int, list]] = None
    ) -> np.ndarray:
        """
        Performs ablation on the input x. If no row_pos/column_pos is specified a random location will be selected.

        :param x: input data
        :param column_pos: Specifies the column index to retain the image block. Either an int to apply the same
                   position to all images in a batch, or a list of ints to apply a different
                   column position per datapoint.
        :param row_pos: Specifies the row index to retain the image block. Either an int to apply the same position to
                        all images in a batch, or a list of ints to apply a different row position per datapoint.
        :return: Data ablated at all locations aside from the specified block. Data is channel extended to indicate to a
                 model if a position is ablated.
        """
        return self.forward(x=x, row_pos=row_pos, column_pos=column_pos)

    def certify(self, preds: np.ndarray, size_to_certify: int) -> np.ndarray:
        """
        Checks if based on the predictions supplied the classifications over the ablated datapoints result in a
        certified prediction against a patch attack of size size_to_certify.

        :param preds: The cumulative predictions of the classifier over the ablation locations.
        :param size_to_certify: The size of the patch to check against.
        :return: Array of bools indicating if a point is certified against the given patch dimensions.
        """
        indices = np.argsort(-preds, axis=1, kind="stable")
        values = np.take_along_axis(np.copy(preds), indices, axis=1)
        margin = values[:, 0] - values[:, 1]

        num_affected_classifications = (size_to_certify + self.ablation_size - 1) ** 2

        certs = margin > 2 * num_affected_classifications
        tie_break_certs = (margin == 2 * num_affected_classifications) & (indices[:, 0] < indices[:, 1])
        return np.logical_or(certs, tie_break_certs)

    def forward(
        self,
        x: np.ndarray,
        column_pos: Optional[Union[int, list]] = None,
        row_pos: Optional[Union[int, list]] = None,
    ) -> np.ndarray:
        """
        Performs ablation on the input x. If no column_pos/row_pos are specified a random location will be selected.

        :param x: input data
        :param row_pos: Specifies the row index to retain the image block. Either an int to apply the same position to
                        all images in a batch, or a list of ints to apply a different row position per datapoint.
        :param column_pos: Specifies the column index to retain the image block. Either an int to apply the same
                           position to all images in a batch, or a list of ints to apply a different
                           column position per datapoint.
        :return: Data ablated at all locations aside from the specified block. Data is channel extended to indicate to a
                 model if a position is ablated.
        """
        if not self.channels_first:
            x = np.transpose(x, (0, 3, 1, 2))

        if row_pos is None:
            row_pos = random.randint(0, x.shape[2])
        if column_pos is None:
            column_pos = random.randint(0, x.shape[3])

        x = np.concatenate([x, 1.0 - x], axis=1)

        if isinstance(row_pos, list) and isinstance(column_pos, list):
            for i, (row, col) in enumerate(zip(row_pos, column_pos)):
                x[i : i + 1] = self.ablate(x[i : i + 1], row_pos=row, column_pos=col)
        elif isinstance(row_pos, int) and isinstance(column_pos, int):
            x = self.ablate(x, row_pos=row_pos, column_pos=column_pos)

        if not self.channels_first:
            x = np.transpose(x, (0, 2, 3, 1))
        return x

    def ablate(self, x: np.ndarray, column_pos: int, row_pos: int) -> np.ndarray:
        """
        Ablates the image only retaining a block starting at (row_pos, column_pos) of height/width "self.ablation_size"

        :param x: input data
        :param row_pos: Specifies the row index where to retain the image block.
        :param column_pos: Specifies the column index where to retain the image block.
        :return: Data ablated at all locations aside from the specified block.
        """
        k = self.ablation_size
        num_of_image_columns = x.shape[3]
        num_of_image_rows = x.shape[2]

        if row_pos + k > x.shape[2] and column_pos + k > x.shape[3]:
            start_of_ablation = column_pos + k - num_of_image_columns
            x[:, :, :, start_of_ablation:column_pos] = 0.0

            start_of_ablation = row_pos + k - num_of_image_rows
            x[:, :, start_of_ablation:row_pos, :] = 0.0

        # only the row wraps
        elif row_pos + k > x.shape[2] and column_pos + k <= x.shape[3]:
            x[:, :, :, :column_pos] = 0.0
            x[:, :, :, column_pos + k :] = 0.0

            start_of_ablation = row_pos + k - num_of_image_rows
            x[:, :, start_of_ablation:row_pos, :] = 0.0

        # only column wraps
        elif row_pos + k <= x.shape[2] and column_pos + k > x.shape[3]:
            start_of_ablation = column_pos + k - num_of_image_columns
            x[:, :, :, start_of_ablation:column_pos] = 0.0

            x[:, :, :row_pos, :] = 0.0
            x[:, :, row_pos + k :, :] = 0.0

        # neither wraps
        elif row_pos + k <= x.shape[2] and column_pos + k <= x.shape[3]:
            x[:, :, :, :column_pos] = 0.0
            x[:, :, :, column_pos + k :] = 0.0

            x[:, :, :row_pos, :] = 0.0
            x[:, :, row_pos + k :, :] = 0.0
        else:
            raise ValueError(f"Ablation failed on row: {row_pos} and column: {column_pos} with size {k}")

        return x



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import pytest

from art.estimators.pytorch import PyTorchEstimator
from art.estimators.speech_recognition.speech_recognizer import SpeechRecognizerMixin
from art.estimators.tensorflow import TensorFlowV2Estimator
from tests.utils import ARTTestFixtureNotImplemented


@pytest.fixture
def audio_sample():
    """
    Create audio sample.
    """
    sample_rate = 16000
    test_input = np.ones((sample_rate)) * 10e3
    return test_input


@pytest.fixture
def audio_data():
    """
    Create audio fixtures of shape (nb_samples=3,) with elements of variable length.
    """
    sample_rate = 16000
    test_input = np.array(
        [
            np.zeros(sample_rate),
            np.ones(sample_rate * 2) * 2e3,
            np.ones(sample_rate * 3) * 3e3,
            np.ones(sample_rate * 3) * 3e3,
        ],
        dtype=object,
    )
    test_target = ["DUMMY"] * test_input.shape[0]
    return test_input, test_target


@pytest.fixture
def audio_batch_padded():
    """
    Create audio fixtures of shape (batch_size=2,) with elements of variable length.
    """
    sample_rate = 16000
    test_input = np.ones((2, sample_rate)) * 2e3
    return test_input


@pytest.fixture
def asr_dummy_estimator(framework):
    def _asr_dummy_estimator(**kwargs):
        asr_dummy = None
        if framework in ("tensorflow2v1", "tensorflow2"):

            class TensorFlowV2ASRDummy(TensorFlowV2Estimator, SpeechRecognizerMixin):
                def get_activations(self):
                    pass

                def predict(self, x):
                    pass

                def loss_gradient(self, x, y, **kwargs):
                    return x

                @property
                def input_shape(self):
                    return None

                def compute_loss(self, x, y, **kwargs):
                    pass

            asr_dummy = TensorFlowV2ASRDummy(channels_first=None, model=None, clip_values=None)
        if framework == "pytorch":

            class PyTorchASRDummy(PyTorchEstimator, SpeechRecognizerMixin):
                def get_activations(self):
                    pass

                def predict(self, x):
                    pass

                def loss_gradient(self, x, y, **kwargs):
                    return x

                @property
                def input_shape(self):
                    return None

                def compute_loss(self, x, y, **kwargs):
                    pass

            asr_dummy = PyTorchASRDummy(channels_first=None, model=None, clip_values=None)
        if asr_dummy is None:
            raise ARTTestFixtureNotImplemented(
                "ASR dummy estimator not available for this framework", asr_dummy_estimator.__name__, framework
            )
        return asr_dummy

    return _asr_dummy_estimator



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging

import numpy as np
import pytest

from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.fixture()
def fix_get_mnist_subset(get_mnist_dataset):
    (x_train_mnist, y_train_mnist), (x_test_mnist, y_test_mnist) = get_mnist_dataset
    n_train = 10
    n_test = 10
    yield x_train_mnist[:n_train], y_train_mnist[:n_train], x_test_mnist[:n_test], y_test_mnist[:n_test]


@pytest.mark.only_with_platform("pytorch")
def test_eot_brightness_pytorch(art_warning, fix_get_mnist_subset):
    try:
        import torch
        from art.preprocessing.expectation_over_transformation.natural_corruptions.brightness.pytorch import (
            EoTBrightnessPyTorch,
        )

        x_train_mnist, y_train_mnist, _, _ = fix_get_mnist_subset
        x_train_mnist = np.transpose(x_train_mnist, (0, 2, 3, 1))  # transpose to NHWC

        nb_samples = 3

        eot = EoTBrightnessPyTorch(nb_samples=nb_samples, delta=(0.2, 0.2), clip_values=(0.0, 1.0))
        x_eot, y_eot = eot.forward(x=torch.from_numpy(x_train_mnist), y=torch.from_numpy(y_train_mnist))

        assert x_eot.shape[0] == nb_samples * x_train_mnist.shape[0]
        assert y_eot.shape[0] == nb_samples * y_train_mnist.shape[0]

        x_eot_expected = np.array(
            [
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.5176471,
                1.0,
                1.0,
                1.0,
                0.6666667,
                0.29803923,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
            ]
        )

        np.testing.assert_almost_equal(x_eot.numpy()[0, 14, :, 0], x_eot_expected)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
def test_eot_brightness_tensorflow_v2(art_warning, fix_get_mnist_subset):
    try:
        from art.preprocessing.expectation_over_transformation.natural_corruptions.brightness.tensorflow import (
            EoTBrightnessTensorFlow,
        )

        x_train_mnist, y_train_mnist, _, _ = fix_get_mnist_subset

        nb_samples = 3

        eot = EoTBrightnessTensorFlow(nb_samples=nb_samples, delta=(0.2, 0.2), clip_values=(0.0, 1.0))
        x_eot, y_eot = eot.forward(x=x_train_mnist, y=y_train_mnist)

        assert x_eot.shape[0] == nb_samples * x_train_mnist.shape[0]
        assert y_eot.shape[0] == nb_samples * y_train_mnist.shape[0]

        x_eot_expected = np.array(
            [
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.5176471,
                1.0,
                1.0,
                1.0,
                0.6666667,
                0.29803923,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
                0.2,
            ]
        )

        np.testing.assert_almost_equal(x_eot.numpy()[0, 14, :, 0], x_eot_expected)

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the task specific estimator for Faster R-CNN v3 in PyTorch.
"""
import logging
from typing import List, Optional, Tuple, Union, TYPE_CHECKING


from art.estimators.object_detection.pytorch_object_detector import PyTorchObjectDetector

if TYPE_CHECKING:
    # pylint: disable=C0412
    import torch
    import torchvision

    from art.utils import CLIP_VALUES_TYPE, PREPROCESSING_TYPE
    from art.defences.preprocessor.preprocessor import Preprocessor
    from art.defences.postprocessor.postprocessor import Postprocessor

logger = logging.getLogger(__name__)


class PyTorchFasterRCNN(PyTorchObjectDetector):
    """
    This class implements a model-specific object detector using Faster R-CNN and PyTorch following the input and output
    formats of torchvision.
    """

    def __init__(
        self,
        model: Optional["torchvision.models.detection.FasterRCNN"] = None,
        input_shape: Tuple[int, ...] = (-1, -1, -1),
        optimizer: Optional["torch.optim.Optimizer"] = None,
        clip_values: Optional["CLIP_VALUES_TYPE"] = None,
        channels_first: Optional[bool] = True,
        preprocessing_defences: Union["Preprocessor", List["Preprocessor"], None] = None,
        postprocessing_defences: Union["Postprocessor", List["Postprocessor"], None] = None,
        preprocessing: "PREPROCESSING_TYPE" = None,
        attack_losses: Tuple[str, ...] = (
            "loss_classifier",
            "loss_box_reg",
            "loss_objectness",
            "loss_rpn_box_reg",
        ),
        device_type: str = "gpu",
    ):
        """
        Initialization.

        :param model: Faster R-CNN model. The output of the model is `List[Dict[str, torch.Tensor]]`, one for
                      each input image. The fields of the Dict are as follows:

                      - boxes [N, 4]: the boxes in [x1, y1, x2, y2] format, with 0 <= x1 < x2 <= W and
                        0 <= y1 < y2 <= H.
                      - labels [N]: the labels for each image.
                      - scores [N]: the scores of each prediction.
        :param input_shape: The shape of one input sample.
        :param optimizer: The optimizer for training the classifier.
        :param clip_values: Tuple of the form `(min, max)` of floats or `np.ndarray` representing the minimum and
               maximum values allowed for features. If floats are provided, these will be used as the range of all
               features. If arrays are provided, each value will be considered the bound for a feature, thus
               the shape of clip values needs to match the total number of features.
        :param channels_first: Set channels first or last.
        :param preprocessing_defences: Preprocessing defence(s) to be applied by the classifier.
        :param postprocessing_defences: Postprocessing defence(s) to be applied by the classifier.
        :param preprocessing: Tuple of the form `(subtrahend, divisor)` of floats or `np.ndarray` of values to be
               used for data preprocessing. The first value will be subtracted from the input. The input will then
               be divided by the second one.
        :param attack_losses: Tuple of any combination of strings of loss components: 'loss_classifier', 'loss_box_reg',
                              'loss_objectness', and 'loss_rpn_box_reg'.
        :param device_type: Type of device to be used for model and tensors, if `cpu` run on CPU, if `gpu` run on GPU
                            if available otherwise run on CPU.
        """
        import torchvision

        if model is None:  # pragma: no cover
            model = torchvision.models.detection.fasterrcnn_resnet50_fpn(
                pretrained=True, progress=True, num_classes=91, pretrained_backbone=True
            )

        super().__init__(
            model=model,
            input_shape=input_shape,
            optimizer=optimizer,
            clip_values=clip_values,
            channels_first=channels_first,
            preprocessing_defences=preprocessing_defences,
            postprocessing_defences=postprocessing_defences,
            preprocessing=preprocessing,
            attack_losses=attack_losses,
            device_type=device_type,
        )



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements Expectation over Transformation preprocessing for image rotation in TensorFlow.
"""
import logging
from typing import Optional, TYPE_CHECKING, Tuple, Union

import numpy as np

from art.preprocessing.expectation_over_transformation.tensorflow import EoTTensorFlowV2

if TYPE_CHECKING:
    # pylint: disable=C0412
    import tensorflow as tf

logger = logging.getLogger(__name__)


class EoTImageRotationTensorFlow(EoTTensorFlowV2):
    """
    This module implements Expectation over Transformation preprocessing for image rotation in TensorFlow.
    """

    params = ["nb_samples", "angles", "clip_values", "label_type"]

    label_types = ["classification"]

    def __init__(
        self,
        nb_samples: int,
        clip_values: Tuple[float, float],
        angles: Union[float, Tuple[float, float]] = 45.0,
        label_type: str = "classification",
        apply_fit: bool = False,
        apply_predict: bool = True,
    ) -> None:
        """
        Create an instance of EoTImageRotationTensorFlow.

        :param nb_samples: Number of random samples per input sample.
        :param clip_values: Tuple of the form `(min, max)` representing the minimum and maximum values allowed
                            for features.
        :param angles: A positive scalar angle in degrees defining the uniform sampling range from negative to
                       positive angles_range.
        :param label_type: String defining the type of labels. Currently supported: `classification`
        :param apply_fit: True if applied during fitting/training.
        :param apply_predict: True if applied during predicting.
        """
        super().__init__(
            apply_fit=apply_fit, apply_predict=apply_predict, nb_samples=nb_samples, clip_values=clip_values
        )

        self.angles = angles
        self.angles_range = (-angles, angles) if isinstance(angles, (int, float)) else angles
        self.label_type = label_type
        self._check_params()

    def _transform(
        self, x: "tf.Tensor", y: Optional["tf.Tensor"], **kwargs
    ) -> Tuple["tf.Tensor", Optional["tf.Tensor"]]:
        """
        Transformation of an input image and its label by randomly sampled rotation.

        :param x: Input samples.
        :param y: Label of the samples `x`.
        :return: Transformed samples and labels.
        """
        import tensorflow as tf
        import tensorflow_addons as tfa

        # pylint: disable=E1120,E1123
        angles = tf.random.uniform(shape=(), minval=self.angles_range[0], maxval=self.angles_range[1])
        angles = angles / 360.0 * 2.0 * np.pi
        x_preprocess = tfa.image.rotate(images=x, angles=angles, interpolation="NEAREST", name=None)
        x_preprocess = tf.clip_by_value(
            t=x_preprocess, clip_value_min=-self.clip_values[0], clip_value_max=self.clip_values[1], name=None
        )
        return x_preprocess, y

    def _check_params(self) -> None:

        # pylint: disable=R0916
        if not isinstance(self.angles, (int, float, tuple)) or (
            isinstance(self.angles, tuple)
            and (
                len(self.angles) != 2
                or not isinstance(self.angles[0], (int, float))
                or not isinstance(self.angles[1], (int, float))
                or self.angles[0] > self.angles[1]
            )
        ):
            raise ValueError("The range of angles must be a float in the range (0.0, 180.0].")

        if self.label_type not in self.label_types:
            raise ValueError(
                f"The input for label_type needs to be one of {self.label_types}, currently receiving "
                f"`{self.label_type}`."
            )



import pytest
import tensorflow as tf
import numpy as np
from tests.attacks.utils import backend_test_classifier_type_check_fail
from art.attacks.evasion.pe_malware_attack import MalwareGDTensorFlow
from art.estimators.classification import TensorFlowV2Classifier

from art.estimators.estimator import BaseEstimator, NeuralNetworkMixin
from art.estimators.classification.classifier import ClassifierMixin

from tests.utils import ARTTestException


@pytest.fixture()
def fix_get_synthetic_data():
    """
    As real malware is hard to share, generate random data of the correct size
    """
    # generate dummy data
    padding_char = 256
    maxlen = 2 ** 20

    # batch of 5 datapoints
    synthetic_data = np.ones((5, maxlen), dtype=np.uint16) * padding_char

    size_of_original_files = [
        int(maxlen * 0.1),  # 1 sample significantly smaller than the maxlen
        int(maxlen * 1.5),  # 1 sample larger then maxlen
        int(maxlen * 0.95),  # 1 sample close to the maximum of the maxlen
        int(maxlen),  # 1 sample at the maxlen
        int(maxlen),
    ]  # 1 sample at the maxlen, this will be assigned a benign label and
    # should not be perturbed by the attack.

    # two class option, later change to binary when ART is generally updated.
    y = np.zeros((5, 1))
    y[0:4] = 1  # assign the first 4 datapoints to be labeled as malware

    # fill in with random numbers
    for i, size in enumerate(size_of_original_files):
        if size > maxlen:
            size = maxlen
        synthetic_data[i, 0:size] = np.random.randint(low=0, high=256, size=(1, size))

    # set the DOS header values:
    synthetic_data[:, 0:2] = [77, 90]
    synthetic_data[:, int(0x3C) : int(0x40)] = 0  # zero the pointer location
    synthetic_data[:, int(0x3C)] = 44  # put in a dummy pointer
    synthetic_data[:, int(0x3C) + 1] = 1  # put in a dummy pointer
    return synthetic_data, y, np.asarray(size_of_original_files)


@pytest.fixture()
def fix_make_dummy_model():
    """
    Create a random model for testing
    """

    def get_prediction_model(param_dic):
        """
        Model going from embeddings to predictions so we can easily optimise the embedding malware embedding.
        Needs to have the same structure as the target model.
        Populated here with "standard" parameters.
        """
        inp = tf.keras.layers.Input(
            shape=(
                param_dic["maxlen"],
                param_dic["embedding_size"],
            )
        )
        filt = tf.keras.layers.Conv1D(
            filters=128,
            kernel_size=500,
            strides=500,
            use_bias=True,
            activation="relu",
            padding="valid",
            name="filt_layer",
        )(inp)
        attn = tf.keras.layers.Conv1D(
            filters=128,
            kernel_size=500,
            strides=500,
            use_bias=True,
            activation="sigmoid",
            padding="valid",
            name="attn_layer",
        )(inp)
        gated = tf.keras.layers.Multiply()([filt, attn])
        feat = tf.keras.layers.GlobalMaxPooling1D()(gated)
        dense = tf.keras.layers.Dense(128, activation="relu", name="dense_layer")(feat)
        output = tf.keras.layers.Dense(1, name="output_layer")(dense)
        return tf.keras.Model(inputs=inp, outputs=output)

    param_dic = {"maxlen": 2 ** 20, "input_dim": 257, "embedding_size": 8}
    prediction_model = get_prediction_model(param_dic)

    model_weights = np.random.normal(loc=0, scale=1.0, size=(257, 8))

    classifier = TensorFlowV2Classifier(
        model=prediction_model,
        nb_classes=2,
        loss_object=tf.keras.losses.BinaryCrossentropy(from_logits=True),
        input_shape=(param_dic["maxlen"], param_dic["embedding_size"]),
    )

    return classifier, model_weights


@pytest.mark.skip_framework("pytorch", "mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2v1")
def test_no_perturbation(art_warning, fix_get_synthetic_data, fix_make_dummy_model):
    """
    Assert that with 0 perturbation the data is unmodified
    """
    try:
        param_dic = {"maxlen": 2 ** 20, "input_dim": 257, "embedding_size": 8}
        # First check: with no perturbation the malware of sufficient size, and benign files, should be unperturbed
        attack = MalwareGDTensorFlow(
            classifier=fix_make_dummy_model[0], embedding_weights=fix_make_dummy_model[1], l_0=0, param_dic=param_dic
        )

        attack.l_0 = 0
        x = np.copy(fix_get_synthetic_data[0])
        y = np.copy(fix_get_synthetic_data[1])
        size_of_files = np.copy(fix_get_synthetic_data[2])

        adv_x, adv_y, adv_sizes = attack.pull_out_valid_samples(x, y, size_of_files)

        # We should only have 3 files as the following cannot be converted to valid adv samples:
        #   2nd datapoint (file too large to support any modifications)
        #   5th datapoint (benign file)

        assert len(adv_x) == 3

        adv_x = attack.generate(adv_x, adv_y, adv_sizes)

        j = 0
        for i in range(len(x)):
            if i in [0, 2, 3]:
                assert np.array_equal(adv_x[j], x[i])
                j += 1
            else:
                assert np.array_equal(x[i], fix_get_synthetic_data[0][i])
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2v1")
def test_append_attack(art_warning, fix_get_synthetic_data, fix_make_dummy_model):
    """
    Check append attack wih a given l0 budget
    """
    try:
        param_dic = {"maxlen": 2 ** 20, "input_dim": 257, "embedding_size": 8}
        l0_budget = 1250
        attack = MalwareGDTensorFlow(
            classifier=fix_make_dummy_model[0],
            embedding_weights=fix_make_dummy_model[1],
            l_0=l0_budget,
            param_dic=param_dic,
        )

        x = np.copy(fix_get_synthetic_data[0])
        y = np.copy(fix_get_synthetic_data[1])
        size_of_files = np.copy(fix_get_synthetic_data[2])

        adv_x, adv_y, adv_sizes = attack.pull_out_valid_samples(x, y, size_of_files)

        # We should only have 2 files as the following cannot be converted to valid adv samples:
        #   2nd datapoint (file too large to support any modifications)
        #   4th datapoint (file to large to support append attacks)
        #   5th datapoint (benign file)

        assert len(adv_x) == 2

        adv_x = attack.generate(adv_x, adv_y, adv_sizes)

        j = 0
        for i, size in enumerate(fix_get_synthetic_data[2]):
            if i in [0, 2]:
                assert np.array_equal(adv_x[j, :size], fix_get_synthetic_data[0][i, :size])
                assert not np.array_equal(
                    adv_x[j, size : size + l0_budget], fix_get_synthetic_data[0][i, size : size + l0_budget]
                )
                assert np.array_equal(adv_x[j, size + l0_budget :], fix_get_synthetic_data[0][i, size + l0_budget :])
                j += 1
            else:
                assert np.array_equal(x[i], fix_get_synthetic_data[0][i])
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2v1")
def test_slack_attacks(art_warning, fix_get_synthetic_data, fix_make_dummy_model):
    """
    Testing modification of certain regions in the PE file
    """
    try:
        # Third check: Slack insertion attacks.

        def generate_synthetic_slack_regions(size):
            """
            Generate 4 slack regions per sample, each of size 250.
            """

            batch_of_slack_starts = []
            batch_of_slack_sizes = []

            for _ in range(5):
                size_of_slack = []
                start_of_slack = []
                start = 0
                for _ in range(4):
                    start += 1000
                    start_of_slack.append(start)
                    size_of_slack.append(size)
                batch_of_slack_starts.append(start_of_slack)
                batch_of_slack_sizes.append(size_of_slack)
            return batch_of_slack_starts, batch_of_slack_sizes

        param_dic = {"maxlen": 2 ** 20, "input_dim": 257, "embedding_size": 8}
        # First check: with no perturbation the malware of sufficient size, and benign files, should be unperturbed
        attack = MalwareGDTensorFlow(
            classifier=fix_make_dummy_model[0], embedding_weights=fix_make_dummy_model[1], l_0=0, param_dic=param_dic
        )

        l0_budget = 1250
        attack.l_0 = l0_budget
        x = np.copy(fix_get_synthetic_data[0])
        y = np.copy(fix_get_synthetic_data[1])
        size_of_files = np.copy(fix_get_synthetic_data[2])

        batch_of_section_starts, batch_of_section_sizes = generate_synthetic_slack_regions(size=250)
        adv_x, adv_y, adv_sizes, batch_of_section_starts, batch_of_section_sizes = attack.pull_out_valid_samples(
            x,
            y,
            sample_sizes=size_of_files,
            perturb_starts=batch_of_section_starts,
            perturb_sizes=batch_of_section_sizes,
        )

        # We should only have 2 files as the following cannot be converted to valid adv samples:
        #   2nd datapoint (file too large to support any modifications)
        #   4th datapoint (attack requires appending 250 bytes to end of file which this datapoint cannot support)
        #   5th datapoint (benign file)
        assert len(adv_x) == 2

        adv_x = attack.generate(
            adv_x, adv_y, adv_sizes, perturb_sizes=batch_of_section_sizes, perturb_starts=batch_of_section_starts
        )

        j = 0
        for i, size in enumerate(fix_get_synthetic_data[2]):
            if i in [0, 2]:
                slack_starts = batch_of_section_starts[j]
                slack_sizes = batch_of_section_sizes[j]
                beginning_pos = 0
                total_perturbation = 0
                for slack_start, slack_size in zip(slack_starts, slack_sizes):
                    assert np.array_equal(
                        adv_x[j, beginning_pos:slack_start], fix_get_synthetic_data[0][i, beginning_pos:slack_start]
                    )

                    assert not np.array_equal(
                        adv_x[j, slack_start : slack_start + slack_size],
                        fix_get_synthetic_data[0][i, slack_start : slack_start + slack_size],
                    )
                    beginning_pos = slack_start + slack_size
                    total_perturbation += slack_size

                # from the last slack region to end of file
                assert np.array_equal(adv_x[j, beginning_pos:size], fix_get_synthetic_data[0][i, beginning_pos:size])
                remaining_perturbation = l0_budget - total_perturbation
                assert remaining_perturbation == 250

                # append portion of the attack has been conducted
                assert not np.array_equal(
                    adv_x[j, size : size + remaining_perturbation],
                    fix_get_synthetic_data[0][i, size : size + remaining_perturbation],
                )

                # from end of append to end of datapoint
                assert np.array_equal(
                    adv_x[j, size + remaining_perturbation :],
                    fix_get_synthetic_data[0][i, size + remaining_perturbation :],
                )

                j += 1
            else:
                assert np.array_equal(x[i], fix_get_synthetic_data[0][i])
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2v1")
def test_large_append(art_warning, fix_get_synthetic_data, fix_make_dummy_model):
    """
    Testing with very large perturbation budgets
    """
    # Fourth check append large perturbation
    try:
        param_dic = {"maxlen": 2 ** 20, "input_dim": 257, "embedding_size": 8}
        # First check: with no perturbation the malware of sufficient size, and benign files, should be unperturbed
        l0_budget = int(((2 ** 20) * 0.2))
        attack = MalwareGDTensorFlow(
            classifier=fix_make_dummy_model[0],
            embedding_weights=fix_make_dummy_model[1],
            l_0=l0_budget,
            param_dic=param_dic,
        )

        x = np.copy(fix_get_synthetic_data[0])
        y = np.copy(fix_get_synthetic_data[1])
        size_of_files = np.copy(fix_get_synthetic_data[2])

        adv_x, adv_y, adv_sizes = attack.pull_out_valid_samples(x, y, sample_sizes=size_of_files)
        # We should only have one datapoint that can support an append perturbation of this size
        assert len(adv_x) == 1

        adv_x = attack.generate(adv_x, adv_y, adv_sizes)

        j = 0
        for i, size in enumerate(fix_get_synthetic_data[2]):
            if i == 0:
                assert np.array_equal(adv_x[j, :size], fix_get_synthetic_data[0][i, :size])
                assert not np.array_equal(
                    adv_x[j, size : size + l0_budget], fix_get_synthetic_data[0][i, size : size + l0_budget]
                )
                assert np.array_equal(adv_x[j, size + l0_budget :], fix_get_synthetic_data[0][i, size + l0_budget :])
            else:
                assert np.array_equal(x[i], fix_get_synthetic_data[0][i])
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2v1")
def test_dos_header_attack(art_warning, fix_get_synthetic_data, fix_make_dummy_model):
    """
    Test the DOS header attack modifies the correct regions
    """
    # 5th check: DOS header attack
    try:
        param_dic = {"maxlen": 2 ** 20, "input_dim": 257, "embedding_size": 8}
        # First check: with no perturbation the malware of sufficient size, and benign files, should be unperturbed
        l0_budget = 290
        attack = MalwareGDTensorFlow(
            classifier=fix_make_dummy_model[0],
            embedding_weights=fix_make_dummy_model[1],
            l_0=l0_budget,
            param_dic=param_dic,
        )
        x = np.copy(fix_get_synthetic_data[0])
        y = np.copy(fix_get_synthetic_data[1])
        size_of_files = np.copy(fix_get_synthetic_data[2])

        dos_starts, dos_sizes = attack.get_dos_locations(x)

        adv_x, adv_y, adv_sizes, batch_of_section_starts, batch_of_section_sizes = attack.pull_out_valid_samples(
            x, y, sample_sizes=size_of_files, perturb_starts=dos_starts, perturb_sizes=dos_sizes
        )

        # should have 3 files. Samples which are excluded are:
        #   2nd datapoint (file to large to support any modifications)
        #   5th datapoint (benign file)

        assert len(adv_x) == 3

        adv_x = attack.generate(
            adv_x, adv_y, adv_sizes, perturb_sizes=batch_of_section_sizes, perturb_starts=batch_of_section_starts
        )

        j = 0
        for i in range(len(fix_get_synthetic_data[2])):
            if i in [0, 2, 3]:
                assert np.array_equal(adv_x[j, 0:2], [77, 90])

                # we should have 58 bytes that were perturbed between the magic number and the pointer
                assert not np.array_equal(adv_x[j, 2 : int(0x3C)], fix_get_synthetic_data[0][i, 2 : int(0x3C)])

                # dummy pointer should be unchanged
                assert np.array_equal(adv_x[j, int(0x3C) : int(0x3C) + 4], [44, 1, 0, 0])

                # the remaining perturbation 290 - 58 = 232 is in the rest of the DOS header
                assert not np.array_equal(
                    adv_x[j, int(0x3C) + 4 : int(0x3C) + 4 + 232],
                    fix_get_synthetic_data[0][i, int(0x3C) + 4 : int(0x3C) + 4 + 232],
                )

                # rest of the file is unchanged
                assert np.array_equal(
                    adv_x[j, int(0x3C) + 4 + 232 :], fix_get_synthetic_data[0][i, int(0x3C) + 4 + 232 :]
                )
                j += 1
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2v1")
def test_no_auto_append(art_warning, fix_get_synthetic_data, fix_make_dummy_model):
    """
    Verify behaviour when not spilling extra perturbation into an append attack
    """
    try:

        def generate_synthetic_slack_regions(size):
            """
            Generate 4 slack regions per sample, each of size 250.
            """

            batch_of_slack_starts = []
            batch_of_slack_sizes = []

            for _ in range(5):
                size_of_slack = []
                start_of_slack = []
                start = 0
                for _ in range(4):
                    start += 1000
                    start_of_slack.append(start)
                    size_of_slack.append(size)
                batch_of_slack_starts.append(start_of_slack)
                batch_of_slack_sizes.append(size_of_slack)
            return batch_of_slack_starts, batch_of_slack_sizes

        # 6th check: Do not automatically append extra perturbation
        l0_budget = 1250
        param_dic = {"maxlen": 2 ** 20, "input_dim": 257, "embedding_size": 8}
        attack = MalwareGDTensorFlow(
            classifier=fix_make_dummy_model[0],
            embedding_weights=fix_make_dummy_model[1],
            l_0=l0_budget,
            param_dic=param_dic,
        )
        x = np.copy(fix_get_synthetic_data[0])
        y = np.copy(fix_get_synthetic_data[1])
        size_of_files = np.copy(fix_get_synthetic_data[2])

        attack.l_0 = l0_budget

        batch_of_section_starts, batch_of_section_sizes = generate_synthetic_slack_regions(size=250)
        adv_x, adv_y, adv_sizes, batch_of_section_starts, batch_of_section_sizes = attack.pull_out_valid_samples(
            x,
            y,
            sample_sizes=size_of_files,
            perturb_starts=batch_of_section_starts,
            perturb_sizes=batch_of_section_sizes,
            automatically_append=False,
        )
        # should have 3 samples.
        # 2nd datapoint cannot support any modification
        # 5th is a benign sample
        assert len(adv_x) == 3

        adv_x = attack.generate(
            adv_x,
            adv_y,
            adv_sizes,
            automatically_append=False,
            perturb_sizes=batch_of_section_sizes,
            perturb_starts=batch_of_section_starts,
        )

        j = 0
        for i in range(len(fix_get_synthetic_data[0])):
            if i in [0, 2, 3]:
                slack_starts = batch_of_section_starts[j]
                slack_sizes = batch_of_section_sizes[j]
                beginning_pos = 0
                total_perturbation = 0
                for slack_start, slack_size in zip(slack_starts, slack_sizes):
                    assert np.array_equal(
                        adv_x[j, beginning_pos:slack_start], fix_get_synthetic_data[0][i, beginning_pos:slack_start]
                    )

                    assert not np.array_equal(
                        adv_x[j, slack_start : slack_start + slack_size],
                        fix_get_synthetic_data[0][i, slack_start : slack_start + slack_size],
                    )
                    # set the position to the end of the slack region
                    beginning_pos = slack_start + slack_size
                    total_perturbation += slack_size
                # from end of final inserted perturbation to EOF.
                assert np.array_equal(adv_x[j, beginning_pos:], fix_get_synthetic_data[0][i, beginning_pos:])
                j += 1
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2v1")
def test_do_not_check_for_valid(art_warning, fix_get_synthetic_data, fix_make_dummy_model):
    """
    No checking for valid data. Expect a mixed adversarial/normal data to be returned.
    """
    try:
        l0_budget = 1250
        param_dic = {"maxlen": 2 ** 20, "input_dim": 257, "embedding_size": 8}
        attack = MalwareGDTensorFlow(
            classifier=fix_make_dummy_model[0],
            embedding_weights=fix_make_dummy_model[1],
            l_0=l0_budget,
            param_dic=param_dic,
        )
        x = np.copy(fix_get_synthetic_data[0])
        y = np.copy(fix_get_synthetic_data[1])
        size_of_files = np.copy(fix_get_synthetic_data[2])

        attack.l_0 = l0_budget

        adv_x = attack.generate(x, y, size_of_files, verify_input_data=False)
        assert len(adv_x) == 5

        # We expect 2 files to have been made adversarial the following cannot be converted to valid adv samples:
        #   2nd datapoint (file too large to support any modifications)
        #   4th datapoint (file to large to support append attacks)
        #   5th datapoint (benign file)
        for i, size in enumerate(size_of_files):
            if i in [0, 2]:
                assert np.array_equal(adv_x[i, :size], fix_get_synthetic_data[0][i, :size])
                assert not np.array_equal(
                    adv_x[i, size : size + l0_budget], fix_get_synthetic_data[0][i, size : size + l0_budget]
                )
                assert np.array_equal(adv_x[i, size + l0_budget :], fix_get_synthetic_data[0][i, size + l0_budget :])
            else:
                assert np.array_equal(adv_x[i], fix_get_synthetic_data[0][i])
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2v1")
def test_check_params(art_warning, image_dl_estimator_for_attack):
    try:
        classifier = image_dl_estimator_for_attack(MalwareGDTensorFlow)

        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(classifier, param_dic=[1, 2, 3], embedding_weights=np.array([1, 2, 3]))

        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(classifier, param_dic={"test": 1}, embedding_weights=[1, 2, 3])

        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(classifier, param_dic={"test": 1}, embedding_weights=np.array([1, 2, 3]), l_0="1")
        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(classifier, param_dic={"test": 1}, embedding_weights=np.array([1, 2, 3]), l_0=-1)

        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(classifier, param_dic={"test": 1}, embedding_weights=np.array([1, 2, 3]), l_r="1")
        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(classifier, param_dic={"test": 1}, embedding_weights=np.array([1, 2, 3]), l_r=-1)

        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(
                classifier, param_dic={"test": 1}, embedding_weights=np.array([1, 2, 3]), use_sign="true"
            )

        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(
                classifier, param_dic={"test": 1}, embedding_weights=np.array([1, 2, 3]), num_of_iterations=1.0
            )
        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(
                classifier, param_dic={"test": 1}, embedding_weights=np.array([1, 2, 3]), num_of_iterations=-1
            )

        with pytest.raises(ValueError):
            _ = MalwareGDTensorFlow(
                classifier, param_dic={"test": 1}, embedding_weights=np.array([1, 2, 3]), verbose="true"
            )

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.framework_agnostic
def test_classifier_type_check_fail(art_warning, fix_make_dummy_model):
    try:
        param_dic = {"maxlen": 2 ** 20, "input_dim": 257, "embedding_size": 8}
        backend_test_classifier_type_check_fail(
            MalwareGDTensorFlow,
            [BaseEstimator, NeuralNetworkMixin, ClassifierMixin],
            classifier=None,
            param_dic=param_dic,
            embedding_weights=fix_make_dummy_model[1],
        )

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import unittest
import keras.backend as k
import numpy as np

from art.attacks.evasion import GRAPHITEBlackbox, GRAPHITEWhiteboxPyTorch
from art.estimators.estimator import BaseEstimator
from art.estimators.classification import ClassifierMixin

from tests.utils import TestBase
from tests.utils import get_image_classifier_tf, get_image_classifier_kr, get_image_classifier_pt

from tests.utils import master_seed
from tests.attacks.utils import backend_test_classifier_type_check_fail

logger = logging.getLogger(__name__)


class TestGRAPHITE(TestBase):
    """
    A unittest class for testing the GRAPHITE attack.
    """

    @classmethod
    def setUpClass(cls):
        master_seed(seed=1234, set_tensorflow=True, set_torch=True)
        super().setUpClass()

        cls.n_test = 2
        cls.x_test_init_mnist = cls.x_test_mnist[1 : cls.n_test + 1]
        cls.y_test_init_mnist = cls.y_test_mnist[1 : cls.n_test + 1]
        cls.x_test_mnist = cls.x_test_mnist[0 : cls.n_test]
        cls.y_test_mnist = cls.y_test_mnist[0 : cls.n_test]

    def setUp(self):
        master_seed(seed=1234, set_tensorflow=True, set_torch=True)
        super().setUp()

    def test_3_tensorflow_mnist(self):
        """
        First test with the TensorFlowClassifier.
        :return:
        """
        x_test_original = self.x_test_mnist.copy()

        # Build TensorFlowClassifier
        tfc, sess = get_image_classifier_tf()

        # First attack
        graphite = GRAPHITEBlackbox(
            classifier=tfc,
            noise_size=(28, 28),
            net_size=(28, 28),
            heatmap_mode="Target",
            num_xforms_mask=2,
            num_xforms_boost=10,
            rotation_range=(-5, 5),
            gamma_range=(1.0, 1.1),
            crop_percent_range=(-0.001, 0.001),
            off_x_range=(-0.001, 0.001),
            blur_kernels=[0],
        )
        params = {"y": self.y_test_init_mnist, "x_tar": self.x_test_init_mnist}
        x_test_adv = graphite.generate(self.x_test_mnist, **params)

        self.assertFalse((self.x_test_mnist == x_test_adv).all())
        self.assertTrue((x_test_adv <= 1.0001).all())
        self.assertTrue((x_test_adv >= -0.0001).all())

        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(tfc.predict(x_test_adv), axis=1)
        self.assertTrue((target == y_pred_adv).any())

        # Test the masking
        mask = np.random.binomial(n=1, p=0.5, size=np.prod(self.x_test_mnist.shape))
        mask = mask.reshape(self.x_test_mnist.shape)

        params.update(mask=mask)
        x_test_adv = graphite.generate(self.x_test_mnist, **params)
        mask_diff = (1 - mask) * (x_test_adv - self.x_test_mnist)
        self.assertAlmostEqual(float(np.max(np.abs(mask_diff))), 0.0, delta=0.00001)

        unmask_diff = mask * (x_test_adv - self.x_test_mnist)
        self.assertGreater(float(np.sum(np.abs(unmask_diff))), 0.0)

        # Test passing in points
        pts = np.zeros((4, 3, 1))
        pts[0, :, 0] = np.array([0.05, 0.05, 1])
        pts[1, :, 0] = np.array([0.05, 0.95, 1])
        pts[2, :, 0] = np.array([0.95, 0.05, 1])
        pts[3, :, 0] = np.array([0.95, 0.95, 1])

        params.update(pts=pts)
        x_test_adv = graphite.generate(self.x_test_mnist, **params)
        mask_diff = (1 - mask) * (x_test_adv - self.x_test_mnist)
        self.assertAlmostEqual(float(np.max(np.abs(mask_diff))), 0.0, delta=0.00001)

        unmask_diff = mask * (x_test_adv - self.x_test_mnist)
        self.assertGreater(float(np.sum(np.abs(unmask_diff))), 0.0)

        # Check that x_test has not been modified by attack and classifier
        self.assertAlmostEqual(float(np.max(np.abs(x_test_original - self.x_test_mnist))), 0.0, delta=0.00001)

        # Clean-up session
        if sess is not None:
            sess.close()

    def test_8_keras_mnist(self):
        """
        Second test with the KerasClassifier.
        :return:
        """
        x_test_original = self.x_test_mnist.copy()

        # Build KerasClassifier
        krc = get_image_classifier_kr()

        # First attack
        graphite = GRAPHITEBlackbox(
            classifier=krc,
            noise_size=(28, 28),
            net_size=(28, 28),
            heatmap_mode="Target",
            num_xforms_mask=2,
            num_xforms_boost=10,
            rotation_range=(-5, 5),
            gamma_range=(1.0, 1.1),
            crop_percent_range=(-0.001, 0.001),
            off_x_range=(-0.001, 0.001),
            blur_kernels=[0],
        )
        params = {"y": self.y_test_init_mnist, "x_tar": self.x_test_init_mnist}
        x_test_adv = graphite.generate(self.x_test_mnist, **params)

        self.assertFalse((self.x_test_mnist == x_test_adv).all())
        self.assertTrue((x_test_adv <= 1.0001).all())
        self.assertTrue((x_test_adv >= -0.0001).all())

        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(krc.predict(x_test_adv), axis=1)
        self.assertTrue((target == y_pred_adv).any())

        # Test the masking
        mask = np.random.binomial(n=1, p=0.5, size=np.prod(self.x_test_mnist.shape))
        mask = mask.reshape(self.x_test_mnist.shape)

        params.update(mask=mask)
        x_test_adv = graphite.generate(self.x_test_mnist, **params)
        mask_diff = (1 - mask) * (x_test_adv - self.x_test_mnist)
        self.assertAlmostEqual(float(np.max(np.abs(mask_diff))), 0.0, delta=0.00001)

        unmask_diff = mask * (x_test_adv - self.x_test_mnist)
        self.assertGreater(float(np.sum(np.abs(unmask_diff))), 0.0)

        # Test passing in points
        pts = np.zeros((4, 3, 1))
        pts[0, :, 0] = np.array([0.05, 0.05, 1])
        pts[1, :, 0] = np.array([0.05, 0.95, 1])
        pts[2, :, 0] = np.array([0.95, 0.05, 1])
        pts[3, :, 0] = np.array([0.95, 0.95, 1])

        params.update(pts=pts)
        x_test_adv = graphite.generate(self.x_test_mnist, **params)
        mask_diff = (1 - mask) * (x_test_adv - self.x_test_mnist)
        self.assertAlmostEqual(float(np.max(np.abs(mask_diff))), 0.0, delta=0.00001)

        unmask_diff = mask * (x_test_adv - self.x_test_mnist)
        self.assertGreater(float(np.sum(np.abs(unmask_diff))), 0.0)

        # Check that x_test has not been modified by attack and classifier
        self.assertAlmostEqual(float(np.max(np.abs(x_test_original - self.x_test_mnist))), 0.0, delta=0.00001)

        # Clean-up session
        k.clear_session()

    def test_4_pytorch_classifier(self):
        """
        Third test with the PyTorchClassifier.
        :return:
        """
        x_test = np.transpose(self.x_test_mnist, (0, 3, 1, 2)).astype(np.float32)
        x_test_init = np.transpose(self.x_test_init_mnist, (0, 3, 1, 2)).astype(np.float32)
        x_test_original = x_test.copy()

        # Build PyTorchClassifier
        ptc = get_image_classifier_pt()

        # BLACKBOX
        # First attack
        graphite = GRAPHITEBlackbox(
            classifier=ptc,
            noise_size=(28, 28),
            net_size=(28, 28),
            heatmap_mode="Target",
            num_xforms_mask=2,
            num_xforms_boost=10,
            rotation_range=(-5, 5),
            gamma_range=(1.0, 1.1),
            crop_percent_range=(-0.001, 0.001),
            off_x_range=(-0.001, 0.001),
            blur_kernels=[0],
        )
        params = {"y": self.y_test_init_mnist, "x_tar": x_test_init}
        x_test_adv = graphite.generate(x_test, **params)

        self.assertFalse((x_test == x_test_adv).all())
        self.assertTrue((x_test_adv <= 1.0001).all())
        self.assertTrue((x_test_adv >= -0.0001).all())

        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(ptc.predict(x_test_adv), axis=1)
        self.assertTrue((target == y_pred_adv).any())

        # Test the masking
        mask = np.random.binomial(n=1, p=0.5, size=np.prod(x_test.shape))
        mask = mask.reshape(x_test.shape)

        params.update(mask=mask)
        x_test_adv = graphite.generate(x_test, **params)
        mask_diff = (1 - mask) * (x_test_adv - x_test)
        self.assertAlmostEqual(float(np.max(np.abs(mask_diff))), 0.0, delta=0.00001)

        unmask_diff = mask * (x_test_adv - x_test)
        self.assertGreater(float(np.sum(np.abs(unmask_diff))), 0.0)

        # Test passing in points
        pts = np.zeros((4, 3, 1))
        pts[0, :, 0] = np.array([0.05, 0.05, 1])
        pts[1, :, 0] = np.array([0.05, 0.95, 1])
        pts[2, :, 0] = np.array([0.95, 0.05, 1])
        pts[3, :, 0] = np.array([0.95, 0.95, 1])

        params.update(pts=pts)
        x_test_adv = graphite.generate(x_test, **params)
        mask_diff = (1 - mask) * (x_test_adv - x_test)
        self.assertAlmostEqual(float(np.max(np.abs(mask_diff))), 0.0, delta=0.00001)

        unmask_diff = mask * (x_test_adv - x_test)
        self.assertGreater(float(np.sum(np.abs(unmask_diff))), 0.0)

        # Check that x_test has not been modified by attack and classifier
        self.assertAlmostEqual(float(np.max(np.abs(x_test_original - x_test))), 0.0, delta=0.00001)

        # WHITEBOX
        # First attack
        graphite = GRAPHITEWhiteboxPyTorch(classifier=ptc, net_size=(28, 28), num_xforms=10)
        params = {"y": self.y_test_init_mnist}
        x_test_adv = graphite.generate(x_test, **params)

        self.assertFalse((x_test == x_test_adv).all())
        self.assertTrue((x_test_adv <= 1.0001).all())
        self.assertTrue((x_test_adv >= -0.0001).all())

        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(ptc.predict(x_test_adv), axis=1)
        self.assertTrue((target == y_pred_adv).any())

        # Test the masking
        mask = np.random.binomial(n=1, p=0.5, size=np.prod(x_test.shape))
        mask = mask.reshape(x_test.shape)

        params.update(mask=mask)
        x_test_adv = graphite.generate(x_test, **params)
        mask_diff = (1 - mask) * (x_test_adv - x_test)
        self.assertAlmostEqual(float(np.max(np.abs(mask_diff))), 0.0, delta=0.00001)

        unmask_diff = mask * (x_test_adv - x_test)
        self.assertGreater(float(np.sum(np.abs(unmask_diff))), 0.0)

        # Test passing in points
        pts = np.zeros((4, 3, 1))
        pts[0, :, 0] = np.array([0.05, 0.05, 1])
        pts[1, :, 0] = np.array([0.05, 0.95, 1])
        pts[2, :, 0] = np.array([0.95, 0.05, 1])
        pts[3, :, 0] = np.array([0.95, 0.95, 1])

        params.update(pts=pts)
        x_test_adv = graphite.generate(x_test, **params)
        mask_diff = (1 - mask) * (x_test_adv - x_test)
        self.assertAlmostEqual(float(np.max(np.abs(mask_diff))), 0.0, delta=0.00001)

        unmask_diff = mask * (x_test_adv - x_test)
        self.assertGreater(float(np.sum(np.abs(unmask_diff))), 0.0)

        # Check that x_test has not been modified by attack and classifier
        self.assertAlmostEqual(float(np.max(np.abs(x_test_original - x_test))), 0.0, delta=0.00001)

    def test_check_params(self):

        ptc = get_image_classifier_pt(from_logits=True)

        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(1, 1), net_size=(1, 1), heat_patch_size=(2, 2))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), heat_patch_size=(0, 1))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), heat_patch_size=(1.0, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), heat_patch_stride=(0, 1))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), heat_patch_stride=(1.0, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), heatmap_mode="asdf")
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), tr_lo=-1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), tr_lo=1.1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), tr_hi=-1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), tr_hi=1.1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), num_xforms_mask=-1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), num_xforms_mask=1.0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), num_xforms_boost=-1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), num_xforms_boost=1.0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), num_boost_queries=0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), num_boost_queries=1.0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), rotation_range=(-90, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), rotation_range=(90, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), rotation_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), dist_range=(-1, 1))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), dist_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), gamma_range=(0, 1))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), gamma_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), crop_percent_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), off_x_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), off_y_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEBlackbox(ptc, noise_size=(28, 28), net_size=(28, 28), blur_kernels=[-1])

        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), min_tr=-1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), min_tr=1.1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), num_xforms=-1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), num_xforms=1.0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), step_size=0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), step_size=-1)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), first_steps=0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), first_steps=1.0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), steps=0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), steps=1.0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), patch_removal_size=0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), num_patches_to_remove=0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), num_patches_to_remove=1.0)
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), rotation_range=(-90, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), rotation_range=(90, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), rotation_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), dist_range=(-1, 1))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), dist_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), gamma_range=(0, 1))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), gamma_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), crop_percent_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), off_x_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), off_y_range=(1, 0))
        with self.assertRaises(ValueError):
            _ = GRAPHITEWhiteboxPyTorch(ptc, net_size=(28, 28), blur_kernels=[-1])

    def test_1_classifier_type_check_fail(self):
        backend_test_classifier_type_check_fail(
            GRAPHITEBlackbox, [BaseEstimator, ClassifierMixin], noise_size=(28, 28), net_size=(28, 28)
        )
        backend_test_classifier_type_check_fail(
            GRAPHITEWhiteboxPyTorch, [BaseEstimator, ClassifierMixin], net_size=(28, 28)
        )


if __name__ == "__main__":
    unittest.main()



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the Feature Adversaries attack in PyTorch.

| Paper link: https://arxiv.org/abs/1511.05122
"""
import logging
from typing import TYPE_CHECKING, Optional, Tuple, Union

import numpy as np
from tqdm.auto import trange

from art.attacks.attack import EvasionAttack
from art.estimators.estimator import BaseEstimator, NeuralNetworkMixin

if TYPE_CHECKING:
    # pylint: disable=C0412
    import torch
    from torch.optim import Optimizer

    from art.utils import PYTORCH_ESTIMATOR_TYPE

logger = logging.getLogger(__name__)


class FeatureAdversariesPyTorch(EvasionAttack):
    """
    This class represent a Feature Adversaries evasion attack in PyTorch.

    | Paper link: https://arxiv.org/abs/1511.05122
    """

    attack_params = EvasionAttack.attack_params + [
        "delta",
        "optimizer",
        "optimizer_kwargs",
        "lambda_",
        "layer",
        "max_iter",
        "batch_size",
        "step_size",
        "random_start",
        "verbose",
    ]

    _estimator_requirements = (BaseEstimator, NeuralNetworkMixin)

    def __init__(
        self,
        estimator: "PYTORCH_ESTIMATOR_TYPE",
        delta: float,
        optimizer: Optional["Optimizer"] = None,
        optimizer_kwargs: Optional[dict] = None,
        lambda_: float = 0.0,
        layer: Union[int, str, Tuple[int, ...], Tuple[str, ...]] = -1,
        max_iter: int = 100,
        batch_size: int = 32,
        step_size: Optional[Union[int, float]] = None,
        random_start: bool = False,
        verbose: bool = True,
    ):
        """
        Create a :class:`.FeatureAdversariesPyTorch` instance.

        :param estimator: A trained estimator.
        :param delta: The maximum deviation between source and guide images.
        :param optimizer: Optimizer applied to problem constrained only by clip values if defined, if None the
                          Projected Gradient Descent (PGD) optimizer is used.
        :param optimizer_kwargs: Additional optimizer arguments.
        :param lambda_: Regularization parameter of the L-inf soft constraint.
        :param layer: Index or tuple of indices of the representation layer(s).
        :param max_iter: The maximum number of iterations.
        :param batch_size: Batch size.
        :param step_size: Step size for PGD optimizer.
        :param random_start: Randomly initialize perturbations, when using Projected Gradient Descent variant.
        :param verbose: Show progress bars.
        """
        super().__init__(estimator=estimator)

        self.delta = delta
        self.optimizer = optimizer
        self._optimizer_kwargs = {} if optimizer_kwargs is None else optimizer_kwargs
        self.lambda_ = lambda_
        self.layer = layer if isinstance(layer, tuple) else (layer,)
        self.batch_size = batch_size
        self.max_iter = max_iter
        self.step_size = step_size
        self.random_start = random_start
        self.verbose = verbose
        self._check_params()

    def _generate_batch(self, x: "torch.Tensor", y: "torch.Tensor") -> "torch.Tensor":
        """
        Generate adversarial batch.

        :param x: Source samples.
        :param y: Guide samples.
        :return: Batch of adversarial examples.
        """
        import torch

        def loss_fn(source_orig, source_adv, guide):
            representation_loss = torch.zeros(size=(source_orig.shape[0],)).to(self.estimator.device)
            for layer_i in self.layer:
                adv_representation = self.estimator.get_activations(source_adv, layer_i, self.batch_size, True)
                guide_representation = self.estimator.get_activations(guide, layer_i, self.batch_size, True)

                dim = tuple(range(1, len(source_adv.shape)))
                soft_constraint = torch.amax(torch.abs(source_adv - source_orig), dim=dim)

                dim = tuple(range(1, len(adv_representation.shape)))
                representation_loss += torch.sum(torch.square(adv_representation - guide_representation), dim=dim)

            loss = torch.mean(representation_loss + self.lambda_ * soft_constraint)
            return loss

        self.estimator.model.eval()

        adv = x.clone().detach().to(self.estimator.device)
        if self.random_start:
            # Starting at a uniformly random point
            adv = adv + torch.empty_like(adv).uniform_(-self.delta, self.delta)
            if self.estimator.clip_values is not None:
                adv = torch.clamp(adv, *self.estimator.clip_values)

        if self.optimizer is None:
            # run a plain-vanilla PGD
            adv.requires_grad = True

            for _ in trange(self.max_iter, desc="Feature Adversaries PyTorch", disable=not self.verbose):
                loss = loss_fn(x, adv, y)
                loss.backward()

                # pgd step
                if adv.grad is not None:
                    adv.data = adv - adv.grad.detach().sign() * self.step_size
                else:
                    raise ValueError("Gradient tensor in PyTorch model is `None`.")
                perturbation = torch.clamp(adv.detach() - x.detach(), -self.delta, self.delta)
                adv.data = x.detach() + perturbation
                if self.estimator.clip_values is not None:
                    adv.data = torch.clamp(adv.detach(), *self.estimator.clip_values)
                adv.grad.zero_()
        else:
            # optimize soft constraint problem with chosen optimizer
            opt = self.optimizer(params=[adv], **self._optimizer_kwargs)  # type: ignore

            for _ in trange(self.max_iter, desc="Feature Adversaries PyTorch", disable=not self.verbose):
                adv.requires_grad = True
                # clip adversarial within L-Inf delta-ball
                adv.data = x.detach() + torch.clamp(adv.detach() - x.detach(), -self.delta, self.delta)

                def closure():
                    if torch.is_grad_enabled():
                        opt.zero_grad()
                    loss = loss_fn(x, adv, y)
                    if loss.requires_grad:
                        loss.backward()
                    return loss

                opt.step(closure)

                if self.estimator.clip_values is not None:
                    adv.data = torch.clamp(adv.detach(), *self.estimator.clip_values)
        return adv.detach().cpu()

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """
        Generate adversarial samples and return them in an array.

        :param x: Source samples.
        :param y: Guide samples.
        :return: Adversarial examples.
        """
        import torch

        if y is None:
            raise ValueError("The value of guide `y` cannot be None. Please provide a `np.ndarray` of guide inputs.")
        if x.shape != y.shape:
            raise ValueError("The shape of source `x` and guide `y` must be of same shape.")
        if x.shape[1:] != self.estimator.input_shape:  # pragma: no cover
            raise ValueError("Source and guide inputs must match `input_shape` of estimator.")

        nb_samples = x.shape[0]

        x_adversarial = [None] * nb_samples

        nb_batches = int(np.ceil(nb_samples / float(self.batch_size)))
        for m in range(nb_batches):
            # batch indices
            begin, end = m * self.batch_size, min((m + 1) * self.batch_size, nb_samples)

            # create batch of adversarial examples
            source_batch = torch.tensor(x[begin:end]).to(self.estimator.device)
            guide_batch = torch.tensor(y[begin:end]).to(self.estimator.device)
            x_adversarial[begin:end] = self._generate_batch(source_batch, guide_batch).numpy()
        return np.array(x_adversarial, dtype=x.dtype)

    def _check_params(self) -> None:
        """
        Apply attack-specific checks.
        """
        if not isinstance(self.delta, float):
            raise ValueError("The value of delta must be of type float.")
        if self.delta <= 0:
            raise ValueError("The maximum deviation value delta has to be positive.")

        if not isinstance(self.lambda_, float):
            raise ValueError("The value of lambda_ must be of type float.")
        if self.lambda_ < 0.0:
            raise ValueError("The regularization parameter `lambda_` has to be non-negative.")

        if not isinstance(self.layer[0], (int, str)):
            raise ValueError("The value of the representation layer must be integer or string.")

        if not isinstance(self.max_iter, int):
            raise ValueError("The value of max_iter must be of type int.")
        if self.max_iter <= 0:
            raise ValueError("The maximum number of iterations has to be a positive.")

        if self.batch_size <= 0:
            raise ValueError("The batch size has to be positive.")

        if self.optimizer is None and self.step_size is None:
            raise ValueError("The step size cannot be None if optimizer is None.")
        if self.step_size is not None and not isinstance(self.step_size, (int, float)):
            raise ValueError("The value of step_size must be of type int or float.")
        if self.step_size is not None and self.step_size <= 0:
            raise ValueError("The step size has to be a positive.")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the abstract estimator `JaxEstimator` for Jax models.
"""
import logging

import numpy as np

from art.estimators.estimator import BaseEstimator, LossGradientsMixin, NeuralNetworkMixin


logger = logging.getLogger(__name__)


class JaxEstimator(NeuralNetworkMixin, LossGradientsMixin, BaseEstimator):
    """
    Estimator class for Jax models.
    """

    estimator_params = BaseEstimator.estimator_params + NeuralNetworkMixin.estimator_params

    def __init__(self, **kwargs) -> None:
        """
        Estimator class for Jax models.

        :param channels_first: Set channels first or last.
        :param clip_values: Tuple of the form `(min, max)` of floats or `np.ndarray` representing the minimum and
               maximum values allowed for features. If floats are provided, these will be used as the range of all
               features. If arrays are provided, each value will be considered the bound for a feature, thus
               the shape of clip values needs to match the total number of features.
        :param preprocessing_defences: Preprocessing defence(s) to be applied by the classifier.
        :param postprocessing_defences: Postprocessing defence(s) to be applied by the classifier.
        :param preprocessing: Tuple of the form `(subtrahend, divisor)` of floats or `np.ndarray` of values to be
               used for data preprocessing. The first value will be subtracted from the input. The input will then
               be divided by the second one.
        :param device_type: Type of device on which the classifier is run, either `gpu` or `cpu`.
        """
        super().__init__(**kwargs)

        JaxEstimator._check_params(self)

    def predict(self, x: np.ndarray, batch_size: int = 128, **kwargs):
        """
        Perform prediction of the neural network for samples `x`.

        :param x: Samples of shape (nb_samples, nb_features) or (nb_samples, nb_pixels_1, nb_pixels_2,
                  nb_channels) or (nb_samples, nb_channels, nb_pixels_1, nb_pixels_2).
        :param batch_size: Batch size.
        :return: Predictions.
        :rtype: Format as expected by the `model`
        """
        return NeuralNetworkMixin.predict(self, x, batch_size=batch_size, **kwargs)

    def fit(self, x: np.ndarray, y, batch_size: int = 128, nb_epochs: int = 20, **kwargs) -> None:
        """
        Fit the model of the estimator on the training data `x` and `y`.

        :param x: Samples of shape (nb_samples, nb_features) or (nb_samples, nb_pixels_1, nb_pixels_2,
                  nb_channels) or (nb_samples, nb_channels, nb_pixels_1, nb_pixels_2).
        :param y: Target values.
        :type y: Format as expected by the `model`
        :param batch_size: Batch size.
        :param nb_epochs: Number of training epochs.
        """
        NeuralNetworkMixin.fit(self, x, y, batch_size=batch_size, nb_epochs=nb_epochs, **kwargs)

    def set_params(self, **kwargs) -> None:
        """
        Take a dictionary of parameters and apply checks before setting them as attributes.

        :param kwargs: A dictionary of attributes.
        """
        super().set_params(**kwargs)
        self._check_params()

    def _check_params(self) -> None:
        super()._check_params()



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from art.config import ART_NUMPY_DTYPE
from art.defences.preprocessor import CutoutTensorFlowV2
from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.fixture(params=[1, 3], ids=["grayscale", "RGB"])
def image_batch(request, channels_first):
    """
    Image fixtures of shape NHWC and NCHW.
    """
    channels = request.param

    if channels_first:
        data_shape = (2, channels, 12, 8)
    else:
        data_shape = (2, 12, 8, channels)
    return (255 * np.ones(data_shape)).astype(ART_NUMPY_DTYPE)


@pytest.fixture(params=[1, 3], ids=["grayscale", "RGB"])
def video_batch(request, channels_first):
    """
    Video fixtures of shape NFHWC and NCFHW.
    """
    channels = request.param

    if channels_first:
        data_shape = (2, 2, channels, 12, 8)
    else:
        data_shape = (2, 2, 12, 8, channels)
    return (255 * np.ones(data_shape)).astype(ART_NUMPY_DTYPE)


@pytest.fixture(params=[1, 3], ids=["grayscale", "RGB"])
def empty_image(request, channels_first):
    """
    Empty image fixtures of shape NHWC and NCHW.
    """
    channels = request.param

    if channels_first:
        data_shape = (2, channels, 12, 8)
    else:
        data_shape = (2, 12, 8, channels)
    return np.zeros(data_shape).astype(ART_NUMPY_DTYPE)


@pytest.mark.only_with_platform("tensorflow2")
@pytest.mark.parametrize("length", [2, 4])
@pytest.mark.parametrize("channels_first", [True, False])
def test_cutout_image_data(art_warning, image_batch, length, channels_first):
    try:
        cutout = CutoutTensorFlowV2(length=length, channels_first=channels_first)
        count = np.not_equal(cutout(image_batch)[0], image_batch).sum()

        n = image_batch.shape[0]
        if channels_first:
            channels = image_batch.shape[1]
        else:
            channels = image_batch.shape[-1]
        assert count <= n * channels * length * length
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
@pytest.mark.parametrize("length", [4])
@pytest.mark.parametrize("channels_first", [True, False])
def test_cutout_video_data(art_warning, video_batch, length, channels_first):
    try:
        cutout = CutoutTensorFlowV2(length=length, channels_first=channels_first)
        count = np.not_equal(cutout(video_batch)[0], video_batch).sum()

        n = video_batch.shape[0]
        frames = video_batch.shape[1]
        if channels_first:
            channels = video_batch.shape[2]
        else:
            channels = video_batch.shape[-1]
        assert count <= n * frames * channels * length * length
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
@pytest.mark.parametrize("length", [4])
@pytest.mark.parametrize("channels_first", [True])
def test_cutout_empty_data(art_warning, empty_image, length, channels_first):
    try:
        cutout = CutoutTensorFlowV2(length=length, channels_first=channels_first)
        assert_array_equal(cutout(empty_image)[0], empty_image)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
def test_non_image_data_error(art_warning, tabular_batch):
    try:
        test_input = tabular_batch
        cutout = CutoutTensorFlowV2(length=8, channels_first=True)

        exc_msg = "Unrecognized input dimension. Cutout can only be applied to image and video data."
        with pytest.raises(ValueError, match=exc_msg):
            cutout(test_input)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
def test_check_params(art_warning):
    try:
        with pytest.raises(ValueError):
            _ = CutoutTensorFlowV2(length=-1)

        with pytest.raises(ValueError):
            _ = CutoutTensorFlowV2(length=0)

    except ARTTestException as e:
        art_warning(e)
